"""A PEP 517 build backend for Meltano.

This uses the PEP 517 build backend provided by `poetry-core`, with a slight
modification. The `WheelBuilder.build` method has been patched to allow us to
make arbitrary changes to the source/sdist immediately prior to the
construction of the wheel.

This is used instead of Poetry's unofficial `build.py` approach because Poetry
the assumes that if `build.py` is used, it should build a platform-specific
wheel, rather than a platform-agnostic pure wheel.
"""

from __future__ import annotations

import os
import shutil
import subprocess
from warnings import warn


def build_webapp() -> None:
    """Build the Meltano UI webapp."""
    warning_prefix = "Failed to build Meltano UI webapp:"

    try:
        node_version_check = subprocess.run(
            ("node", "--version"),
            text=True,
            capture_output=True,
        )
        node_version_check.check_returncode()
    except Exception:
        warn(
            f"{warning_prefix} could not execute 'node --version'",
            Warning,
            stacklevel=2,
        )
        return

    try:
        subprocess.run(("yarn", "--version")).check_returncode()
    except Exception:
        warn(
            f"{warning_prefix} could not execute 'yarn --version'",
            Warning,
            stacklevel=2,
        )
        return

    if not node_version_check.stdout.startswith("v16"):
        warn(f"{warning_prefix} NodeJS v16 is required", Warning, stacklevel=2)
        return

    # Build static web app
    subprocess.run(("yarn", "install", "--immutable"), cwd="src/webapp")
    subprocess.run(("yarn", "build"), cwd="src/webapp")


def include_webapp() -> None:
    """Copy the built webapp into the source tree."""
    os.makedirs("src/meltano/api/templates", exist_ok=True)
    shutil.copy(
        "src/webapp/dist/index.html",
        "src/meltano/api/templates/webapp.html",
    )
    shutil.copy(
        "src/webapp/dist/index-embed.html",
        "src/meltano/api/templates/embed.html",
    )
    for dst in ("css", "js"):
        shutil.rmtree(f"src/meltano/api/static/{dst}", ignore_errors=True)
        shutil.copytree(
            f"src/webapp/dist/static/{dst}",
            f"src/meltano/api/static/{dst}",
        )


from poetry.core.masonry.builders.wheel import WheelBuilder  # noqa: E402

original_wheel_build = WheelBuilder.build


def custom_build_wheel(*args, **kwargs) -> str:
    """Wheel building via poetry-core with customizations for Meltano.

    Args:
        args: Positional arguments to pass through to `WheelBuilder.build`.
        kwargs: Keyword arguments to pass through to `WheelBuilder.build`.

    Returns:
        The value returned by `WheelBuilder.build(*args, **kwargs)`.
    """
    build_webapp()
    include_webapp()
    return original_wheel_build(*args, **kwargs)


WheelBuilder.build = custom_build_wheel

# Provide the Poetry core PEP 517 API, which will now use the patched `WheelBuilder`.
from poetry.core.masonry import api  # noqa: F401, E402



from __future__ import annotations

import datetime
import itertools
import logging
import os
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path

import mock
import pytest
import structlog
import yaml

from fixtures.utils import cd, tmp_project
from meltano.core import bundle
from meltano.core.behavior.canonical import Canonical
from meltano.core.elt_context import ELTContextBuilder
from meltano.core.environment_service import EnvironmentService
from meltano.core.job import Job, Payload, State
from meltano.core.job_state import JobState
from meltano.core.logging.formatters import LEVELED_TIMESTAMPED_PRE_CHAIN
from meltano.core.logging.job_logging_service import JobLoggingService
from meltano.core.plugin import PluginType
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_discovery_service import (
    LockedDefinitionService,
    PluginDiscoveryService,
)
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_files import ProjectFiles
from meltano.core.project_init_service import ProjectInitService
from meltano.core.project_plugins_service import PluginAlreadyAddedException
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.schedule_service import ScheduleAlreadyExistsError, ScheduleService
from meltano.core.state_service import StateService
from meltano.core.task_sets_service import TaskSetsService
from meltano.core.utils import merge

current_dir = Path(__file__).parent


@pytest.fixture(scope="class")
def discovery():  # noqa: WPS213
    with open(bundle.root / "discovery.yml") as base:
        discovery = yaml.safe_load(base)

    discovery[PluginType.EXTRACTORS].append(
        {
            "name": "tap-mock",
            "label": "Mock",
            "namespace": "tap_mock",
            "variants": [
                {
                    "name": "meltano",
                    "pip_url": "tap-mock",
                    "executable": "tap-mock",
                    "capabilities": ["discover", "catalog", "state"],
                    "settings": [
                        {"name": "test", "value": "mock"},
                        {"name": "start_date"},
                        {"name": "protected", "protected": True},
                        {"name": "secure", "kind": "password"},
                        {"name": "port", "kind": "integer", "value": 5000},
                        {"name": "list", "kind": "array", "value": []},
                        {
                            "name": "object",
                            "aliases": ["data"],
                            "kind": "object",
                            "value": {"nested": "from_default"},
                        },
                        {"name": "hidden", "kind": "hidden", "value": 42},
                        {"name": "boolean", "kind": "boolean"},
                        {"name": "auth.username"},
                        {"name": "auth.password", "kind": "password"},
                        {
                            "name": "aliased",
                            "kind": "string",
                            "aliases": ["aliased_1", "aliased_2", "aliased_3"],
                        },
                        {"name": "stacked_env_var", "kind": "string"},
                    ],
                    "commands": {
                        "cmd": {
                            "args": "cmd meltano",
                            "description": "a description of cmd",
                        },
                        "cmd-variant": "cmd-variant meltano",
                        "test": {
                            "args": "--test",
                            "description": "Run tests",
                        },
                        "test_extra": {
                            "args": "test_extra",
                            "description": "Run extra tests",
                            "executable": "test-extra",
                        },
                    },
                },
                {
                    "name": "singer-io",
                    "original": True,
                    "deprecated": True,
                    "pip_url": "singer-tap-mock",
                },
            ],
        },
    )

    discovery[PluginType.EXTRACTORS].append(
        {
            "name": "tap-mock-noinstall",
            "label": "Mock",
            "namespace": "tap_mock_noinstall",
            "variants": [
                {
                    "name": "meltano",
                    "executable": "tap-mock-noinstall",
                    "capabilities": ["discover", "catalog", "state"],
                    "settings": [
                        {"name": "test", "value": "mock"},
                        {"name": "start_date"},
                    ],
                },
            ],
        },
    )
    discovery[PluginType.LOADERS].append(
        {
            "name": "target-mock",
            "namespace": "mock",
            "pip_url": "target-mock",
            "settings": [
                {
                    "name": "schema",
                    "env": "MOCKED_SCHEMA",
                },
            ],
        },
    )

    discovery[PluginType.TRANSFORMS].append(
        {
            "name": "tap-mock-transform",
            "namespace": "tap_mock",
            "pip_url": "tap-mock-transform",
            "package_name": "dbt_mock",
        },
    )

    discovery[PluginType.ORCHESTRATORS].append(
        {
            "name": "orchestrator-mock",
            "namespace": "pytest",
            "pip_url": "orchestrator-mock",
        },
    )

    discovery[PluginType.TRANSFORMERS].append(
        {
            "name": "transformer-mock",
            "namespace": "transformer_mock",
            "pip_url": "transformer-mock",
            "requires": {
                "files": [
                    {
                        "name": "files-transformer-mock",
                        "variant": "meltano",
                    },
                ],
            },
        },
    )

    discovery[PluginType.UTILITIES].append(
        {
            "name": "utility-mock",
            "namespace": "utility_mock",
            "pip_url": "utility-mock",
            "executable": "utility-mock",
            "commands": {
                "cmd": {
                    "args": "--option $ENV_VAR_ARG",
                    "description": "description of utility command",
                },
                "alternate-exec": {
                    "args": "--option $ENV_VAR_ARG",
                    "executable": "other-utility",
                },
                "containerized": {
                    "args": "",
                    "container_spec": {
                        "image": "mock-utils/mock",
                        "ports": {
                            "5000": "5000",
                        },
                        "volumes": ["$MELTANO_PROJECT_ROOT/example/:/usr/app/"],
                    },
                },
            },
        },
    )

    discovery[PluginType.MAPPERS].append(
        {
            "name": "mapper-mock",
            "namespace": "mapper_mock",
            "variants": [
                {
                    "name": "meltano",
                    "executable": "mapper-mock-cmd",
                    "pip_url": "mapper-mock",
                    "package_name": "mapper-mock",
                },
                {
                    "name": "alternative",
                    "executable": "mapper-mock-alt",
                    "pip_url": "mapper-mock-alt",
                    "package_name": "mapper-mock-alt",
                },
            ],
        },
    )

    return discovery


@pytest.fixture(scope="class")
def plugin_discovery_service(project, discovery):
    return PluginDiscoveryService(project, discovery=deepcopy(discovery))


@pytest.fixture(scope="class")
def locked_definition_service(project):
    return LockedDefinitionService(project)


@pytest.fixture(scope="class")
def project_init_service(request):
    return ProjectInitService(f"project_{request.node.name}")


@pytest.fixture(scope="class")
def plugin_install_service(project):
    return PluginInstallService(project)


@pytest.fixture(scope="class")
def project_add_service(project):
    return ProjectAddService(project)


@pytest.fixture(scope="class")
def plugin_settings_service_factory(project):
    def _factory(plugin, **kwargs):
        return PluginSettingsService(project, plugin, **kwargs)

    return _factory


@pytest.fixture(scope="class")
def plugin_invoker_factory(project, plugin_settings_service_factory):
    def _factory(plugin, **kwargs):
        return invoker_factory(
            project,
            plugin,
            plugin_settings_service=plugin_settings_service_factory(plugin),
            **kwargs,
        )

    return _factory


@pytest.fixture(scope="class")
def tap(project_add_service):
    try:
        return project_add_service.add(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant="meltano",
        )
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def alternative_tap(project_add_service, tap):
    try:
        return project_add_service.add(
            PluginType.EXTRACTORS,
            "tap-mock--singer-io",
            inherit_from=tap.name,
            variant="singer-io",
        )
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def inherited_tap(project_add_service, tap):
    try:
        return project_add_service.add(
            PluginType.EXTRACTORS,
            "tap-mock-inherited",
            inherit_from=tap.name,
            commands={
                "cmd": "cmd inherited",
                "cmd-inherited": "cmd-inherited",
            },
        )
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def nonpip_tap(project_add_service):
    try:
        return project_add_service.add(
            PluginType.EXTRACTORS,
            "tap-mock-noinstall",
            executable="tap-mock-noinstall",
        )
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def target(project_add_service):
    try:
        return project_add_service.add(PluginType.LOADERS, "target-mock")
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def alternative_target(project_add_service):
    # We don't load the `target` fixture here since this ProjectPlugin should
    # have a BasePlugin parent, not the `target` ProjectPlugin
    try:
        return project_add_service.add(
            PluginType.LOADERS,
            "target-mock-alternative",
            inherit_from="target-mock",
        )
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def dbt(project_add_service):
    try:
        return project_add_service.add(PluginType.TRANSFORMERS, "dbt")
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def transformer(project_add_service: ProjectAddService):
    try:
        return project_add_service.add(PluginType.TRANSFORMERS, "transformer-mock")
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def utility(project_add_service):
    try:
        return project_add_service.add(PluginType.UTILITIES, "utility-mock")
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture(scope="class")
def schedule_service(project):
    return ScheduleService(project)


@pytest.fixture(scope="class")
def task_sets_service(project):
    return TaskSetsService(project)


@pytest.fixture(scope="class")
def elt_schedule(
    project,  # noqa: ARG001
    tap,
    target,
    schedule_service,
):
    try:
        return schedule_service.add_elt(
            None,
            "elt-schedule-mock",
            extractor=tap.name,
            loader=target.name,
            transform="skip",
            interval="@once",
            start_date=datetime.datetime.now(),
        )
    except ScheduleAlreadyExistsError as err:
        return err.schedule


@pytest.fixture(scope="class")
def job_schedule(
    project,  # noqa: ARG001
    tap,  # noqa: ARG001
    target,  # noqa: ARG001
    schedule_service,
):
    try:
        return schedule_service.add(
            "job-schedule-mock",
            "mock-job",
            interval="@once",
        )
    except ScheduleAlreadyExistsError as err:
        return err.schedule


@pytest.fixture()
def environment_service(project):
    service = EnvironmentService(project)
    try:
        yield service
    finally:
        # Remove any added Environments
        for environment in service.list_environments():
            service.remove(environment.name)


@pytest.fixture(scope="class")
def elt_context_builder(project):
    return ELTContextBuilder(project)


@pytest.fixture(scope="class")
def job_logging_service(project):
    return JobLoggingService(project)


@contextmanager
def project_directory(project_init_service):
    project = project_init_service.init(add_discovery=True)
    logging.debug(f"Created new project at {project.root}")

    # empty out the `plugins`
    with project.meltano_update() as meltano:
        meltano.plugins = Canonical()

    ProjectSettingsService(project).set("snowplow.collector_endpoints", "[]")

    # cd into the new project root
    os.chdir(project.root)

    try:
        yield project
    finally:
        Project.deactivate()
        logging.debug(f"Cleaned project at {project.root}")


@pytest.fixture(scope="class")
def project(project_init_service, tmp_path_factory: pytest.TempPathFactory):
    with cd(tmp_path_factory.mktemp("meltano-project-dir")), project_directory(
        project_init_service,
    ) as project:
        yield project


@pytest.fixture()
def project_function(project_init_service, tmp_path: Path):
    with cd(tmp_path), project_directory(project_init_service) as project:
        yield project


@pytest.fixture(scope="class")
def project_files(tmp_path_factory: pytest.TempPathFactory, compatible_copy_tree):
    with cd(tmp_path_factory.mktemp("meltano-project-files")), tmp_project(
        "a_multifile_meltano_project_core",
        current_dir / "multifile_project",
        compatible_copy_tree,
    ) as project:
        yield ProjectFiles(root=project.root, meltano_file_path=project.meltanofile)


@pytest.fixture(scope="class")
def mapper(project_add_service):
    try:
        return project_add_service.add(
            PluginType.MAPPERS,
            "mapper-mock",
            variant="meltano",
            mappings=[
                {
                    "name": "mock-mapping-0",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "author_email",
                                "tap_stream_name": "commits",
                                "type": "MASK-HIDDEN",
                            },
                        ],
                    },
                },
                {
                    "name": "mock-mapping-1",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "given_name",
                                "tap_stream_name": "users",
                                "type": "lowercase",
                            },
                        ],
                    },
                },
            ],
        )
    except PluginAlreadyAddedException as err:
        return err.plugin


def create_state_id(description: str, env: str = "dev") -> str:
    return f"{env}:tap-{description}-to-target-{description}"


@pytest.fixture()
def num_params():
    return 10


@pytest.fixture()
def payloads(num_params):
    mock_payloads_dict = {
        "mock_state_payloads": [
            {
                "singer_state": {
                    f"bookmark-{idx_i}": idx_i + idx_j for idx_j in range(num_params)
                },
            }
            for idx_i in range(num_params)
        ],
        "mock_error_payload": {"error": "failed"},
        "mock_empty_payload": {},
    }
    payloads = namedtuple("payloads", mock_payloads_dict)
    return payloads(**mock_payloads_dict)


@pytest.fixture()
def state_ids(
    num_params,  # noqa: ARG001
):
    state_id_dict = {
        "single_incomplete_state_id": create_state_id("single-incomplete"),
        "single_complete_state_id": create_state_id("single-complete"),
        "multiple_incompletes_state_id": create_state_id("multiple-incompletes"),
        "multiple_completes_state_id": create_state_id("multiple-completes"),
        "single_complete_then_multiple_incompletes_state_id": create_state_id(
            "single-complete-then-multiple-incompletes",
        ),
        "single_incomplete_then_multiple_completes_state_id": create_state_id(
            "single-incomplete-then-multiple-completes",
        ),
    }
    state_ids = namedtuple("state_ids", state_id_dict)
    return state_ids(**state_id_dict)


@pytest.fixture()
def mock_time():
    def _mock_time():
        for idx in itertools.count():  # noqa: WPS526
            yield datetime.datetime(1, 1, 1) + datetime.timedelta(hours=idx)

    return _mock_time()


@pytest.fixture()
def job_args():
    job_args_dict = {
        "complete_job_args": {"state": State.SUCCESS, "payload_flags": Payload.STATE},
        "incomplete_job_args": {
            "state": State.FAIL,
            "payload_flags": Payload.INCOMPLETE_STATE,
        },
    }
    job_args = namedtuple("job_args", job_args_dict)
    return job_args(**job_args_dict)


@pytest.fixture()
def state_ids_with_jobs(state_ids, job_args, payloads, mock_time):
    jobs = {
        state_ids.single_incomplete_state_id: [
            Job(
                job_name=state_ids.single_incomplete_state_id,
                **job_args.incomplete_job_args,
                payload=payloads.mock_state_payloads[0],
            ),
        ],
        state_ids.single_complete_state_id: [
            Job(
                job_name=state_ids.single_complete_state_id,
                payload=payloads.mock_state_payloads[0],
                **job_args.complete_job_args,
            ),
        ],
        state_ids.multiple_incompletes_state_id: [
            Job(
                job_name=state_ids.multiple_incompletes_state_id,
                **job_args.incomplete_job_args,
                payload=payload,
            )
            for payload in payloads.mock_state_payloads
        ],
        state_ids.multiple_completes_state_id: [
            Job(
                job_name=state_ids.multiple_completes_state_id,
                payload=payload,
                **job_args.complete_job_args,
            )
            for payload in payloads.mock_state_payloads
        ],
        state_ids.single_complete_then_multiple_incompletes_state_id: [
            Job(
                job_name=state_ids.single_complete_then_multiple_incompletes_state_id,
                payload=payloads.mock_state_payloads[0],
                **job_args.complete_job_args,
            ),
        ]
        + [
            Job(
                job_name=state_ids.single_complete_then_multiple_incompletes_state_id,
                payload=payload,
                **job_args.incomplete_job_args,
            )
            for payload in payloads.mock_state_payloads[1:]
        ],
        state_ids.single_incomplete_then_multiple_completes_state_id: [
            Job(
                job_name=state_ids.single_incomplete_then_multiple_completes_state_id,
                payload=payloads.mock_state_payloads[0],
                **job_args.incomplete_job_args,
            ),
        ]
        + [
            Job(
                job_name=state_ids.single_incomplete_then_multiple_completes_state_id,
                payload=payload,
                **job_args.complete_job_args,
            )
            for payload in payloads.mock_state_payloads[1:]
        ],
    }
    for job_list in jobs.values():
        for job in job_list:
            job.started_at = next(mock_time)
            job.ended_at = next(mock_time)
    return jobs


@pytest.fixture()
def jobs(state_ids_with_jobs):
    return [job for job_list in state_ids_with_jobs.values() for job in job_list]


@pytest.fixture()
def state_ids_with_expected_states(  # noqa: WPS210
    state_ids,
    payloads,
    state_ids_with_jobs,
):
    final_state = {}
    for state in payloads.mock_state_payloads:
        merge(state, final_state)
    expectations = {
        state_ids.single_complete_state_id: payloads.mock_state_payloads[0],
        state_ids.single_incomplete_state_id: payloads.mock_empty_payload,
    }

    for state_id, job_list in state_ids_with_jobs.items():
        expectations[state_id] = {}
        jobs = defaultdict(list)
        # Get latest complete non-dummy job.
        for job in job_list:
            if job.state == State.STATE_EDIT:
                jobs["dummy"].append(job)
            elif job.payload_flags == Payload.STATE:
                jobs["complete"].append(job)
            elif job.payload_flags == Payload.INCOMPLETE_STATE:
                jobs["incomplete"].append(job)
        latest_job = {
            kind: (
                max(jobs[kind], key=lambda _job: _job.ended_at) if jobs[kind] else None
            )
            for kind in ("complete", "incomplete")
        }
        if latest_job["complete"]:
            expectations[state_id] = merge(
                expectations[state_id],
                latest_job["complete"].payload,
            )

        for job in jobs["incomplete"]:
            if (not latest_job["complete"]) or (
                job.ended_at > latest_job["complete"].ended_at
            ):
                expectations[state_id] = merge(expectations[state_id], job.payload)
        # Get all dummy jobs since latest non-dummy job.
        for job in jobs["dummy"]:
            if (
                not latest_job["complete"]
                or (job.ended_at > latest_job["complete"].ended_at)
            ) and (
                (not latest_job["incomplete"])
                or (job.ended_at > latest_job["incomplete"].ended_at)
            ):
                expectations[state_id] = merge(expectations[state_id], job.payload)
    return list(expectations.items())


@pytest.fixture()
def job_history_session(jobs, session):
    job: Job
    job_names = set()
    for job in jobs:
        job.save(session)
        job_names.add(job.job_name)
    for job_name in job_names:
        job_state = JobState.from_job_history(session, job_name)
        session.add(job_state)
    return session


@pytest.fixture()
def state_service(job_history_session, project):
    return StateService(project, session=job_history_session)


@pytest.fixture()
def project_with_environment(project: Project):
    project.activate_environment("dev")
    project.environment.env["ENVIRONMENT_ENV_VAR"] = "${MELTANO_PROJECT_ROOT}/file.txt"
    try:
        yield project
    finally:
        project.deactivate_environment()


test_log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "test": {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.processors.JSONRenderer(),
            "foreign_pre_chain": LEVELED_TIMESTAMPED_PRE_CHAIN,
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "test",
            "stream": "ext://sys.stderr",
        },
    },
    "loggers": {
        "": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": True,
        },
    },
}


@pytest.fixture()
def use_test_log_config():
    with mock.patch(
        "meltano.core.logging.utils.default_config",
        return_value=test_log_config,
    ) as patched_default_config:
        yield patched_default_config



from __future__ import annotations

import pytest

from meltano.api.models import db


class TestApp:
    @pytest.fixture()
    def session(self):
        # disable the `session` fixture not to override
        # the `db.session`
        pass

    @pytest.mark.usefixtures("app")
    def test_core_registered(self, engine_sessionmaker):
        engine, _ = engine_sessionmaker

        # ensure both the API and the meltano.core
        # are on the same database
        assert engine.url == db.engine.url


class TestAppSMTPDefault:
    DEFAULTS = {
        "MAIL_SERVER": "localhost",
        "MAIL_PORT": 1025,
        "MAIL_DEFAULT_SENDER": '"Meltano" <bot@meltano.com>',
        "MAIL_USE_TLS": False,
        "MAIL_USERNAME": None,
        "MAIL_PASSWORD": None,
        "MAIL_DEBUG": False,
    }

    def test_config_smtp_default(self, app):
        # ensure it is the defaults when not set
        assert self.DEFAULTS.items() <= app.config.items()


class TestAppSMTP:
    ENV = {
        "MELTANO_MAIL_SERVER": "smtp.localdomain",
        "MELTANO_MAIL_PORT": "1337",
        "MELTANO_MAIL_DEFAULT_SENDER": '"Doctor Who" <dr.who@localdomain.com>',
        "MELTANO_MAIL_USE_TLS": "1",
        "MELTANO_MAIL_USERNAME": "username",
        "MELTANO_MAIL_PASSWORD": "password",
        "MELTANO_MAIL_DEBUG": "0",
    }

    EXPECTED = {
        "MAIL_SERVER": "smtp.localdomain",
        "MAIL_PORT": 1337,
        "MAIL_DEFAULT_SENDER": '"Doctor Who" <dr.who@localdomain.com>',
        "MAIL_USE_TLS": True,
        "MAIL_USERNAME": "username",
        "MAIL_PASSWORD": "password",
        "MAIL_DEBUG": False,
    }

    @pytest.fixture()
    def app(self, create_app, monkeypatch):
        # ensure the environment is properly loaded
        for env, value in self.ENV.items():
            monkeypatch.setenv(env, value)

        return create_app()

    def test_config_smtp(self, app):
        assert self.EXPECTED.items() <= app.config.items()



from __future__ import annotations

import pytest
from mock import Mock
from sqlalchemy.exc import OperationalError

from meltano.core.db import connect


class TestConnectionRetries:
    def test_ping_failure(self):
        engine_mock = Mock()

        # check if OperationalError is raised if a connection can't be made
        engine_mock.connect.side_effect = OperationalError(
            "test_error",
            "test_error",
            "test_error",
        )
        with pytest.raises(OperationalError):
            connect(engine=engine_mock, max_retries=3, retry_timeout=0.1)

        assert engine_mock.connect.call_count == 4

        # check reconnect on second call to `engine.connect`
        engine_mock.reset_mock()
        engine_mock.connect.side_effect = [
            OperationalError("test_error", "test_error", "test_error"),
            None,
        ]

        connect(engine=engine_mock, max_retries=3, retry_timeout=0.1)
        assert engine_mock.connect.call_count == 2



from __future__ import annotations

import mock
import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.project_settings_service import (
    ProjectSettingsService,
    SettingValueStore,
)


class TestCliUi:
    @pytest.mark.usefixtures("project")
    def test_ui(self, cli_runner):
        with mock.patch(
            "meltano.cli.ui.APIWorker.start",
        ) as start_api_worker, mock.patch(
            "meltano.cli.ui.UIAvailableWorker.start",
        ) as start_ui_available_worker:
            result = cli_runner.invoke(cli, "ui")
            assert_cli_runner(result)

            assert start_api_worker.called
            assert start_ui_available_worker.called

    def test_ui_setup(self, project, cli_runner, monkeypatch):
        monkeypatch.setenv("MELTANO_UI_SECRET_KEY", "existing_secret_key")

        with mock.patch("meltano.cli.ui.secrets.token_hex", return_value="fake_secret"):
            result = cli_runner.invoke(cli, ["ui", "setup", "meltano.example.com"])

        assert_cli_runner(result)

        assert (
            "Setting 'ui.secret_key' has already been set in the environment"
            in result.stdout
        )

        settings_service = ProjectSettingsService(project)

        assert settings_service.get_with_source("ui.server_name") == (
            "meltano.example.com",
            SettingValueStore.DOTENV,
        )
        assert settings_service.get_with_source("ui.secret_key") == (
            "existing_secret_key",
            SettingValueStore.ENV,
        )
        assert settings_service.get_with_source("ui.password_salt") == (
            "fake_secret",
            SettingValueStore.DOTENV,
        )



from __future__ import annotations

import shutil
from pathlib import Path

import pytest

from meltano.core.project import Project
from meltano.core.state_store import (
    AZStorageStateStoreManager,
    DBStateStoreManager,
    GCSStateStoreManager,
    LocalFilesystemStateStoreManager,
    S3StateStoreManager,
    StateBackend,
    state_store_manager_from_project_settings,
)


@pytest.fixture()
def state_path(tmp_path: Path):
    path = tmp_path / ".meltano" / "state"
    try:
        yield str(path)
    finally:
        shutil.rmtree(path)


def test_state_store_manager_from_project_settings(project: Project, state_path: str):
    project.settings.set(["state_backend", "uri"], StateBackend.SYSTEMDB)
    project.settings.set(["state_backend", "lock_timeout_seconds"], 10)
    db_state_store: DBStateStoreManager = state_store_manager_from_project_settings(
        project.settings,
    )
    assert isinstance(db_state_store, DBStateStoreManager)

    # Local Filesystem
    project.settings.set(["state_backend", "uri"], f"file://{state_path}")
    file_state_store: LocalFilesystemStateStoreManager = (
        state_store_manager_from_project_settings(project.settings)
    )
    assert isinstance(file_state_store, LocalFilesystemStateStoreManager)
    assert file_state_store.state_dir == state_path

    # Azure
    project.settings.set(["state_backend", "uri"], "azure://some_container/some/path")
    project.settings.set(
        ["state_backend", "azure", "connection_string"],
        "SOME_CONNECTION_STRING",
    )
    az_state_store: AZStorageStateStoreManager = (
        state_store_manager_from_project_settings(project.settings)
    )
    assert isinstance(az_state_store, AZStorageStateStoreManager)
    assert az_state_store.container_name == "some_container"
    assert az_state_store.prefix == "/some/path"
    assert az_state_store.connection_string == "SOME_CONNECTION_STRING"

    # GCS
    project.settings.set(["state_backend", "uri"], "gs://some_container/some/path")
    gs_state_store: GCSStateStoreManager = state_store_manager_from_project_settings(
        project.settings,
    )
    assert isinstance(gs_state_store, GCSStateStoreManager)
    assert gs_state_store.bucket == "some_container"
    assert gs_state_store.prefix == "/some/path"

    # AWS S3 (credentials in URI)
    project.settings.set(
        ["state_backend", "uri"],
        "s3://aws_access_key_id:aws_secret_access_key@some_bucket/some/path",
    )
    s3_state_store: S3StateStoreManager = state_store_manager_from_project_settings(
        project.settings,
    )
    assert isinstance(s3_state_store, S3StateStoreManager)
    assert s3_state_store.bucket == "some_bucket"
    assert s3_state_store.prefix == "/some/path"
    assert s3_state_store.aws_access_key_id == "aws_access_key_id"  # noqa: S105
    assert s3_state_store.aws_secret_access_key == "aws_secret_access_key"  # noqa: S105

    # AWS S3 (credentials provided directly)
    project.settings.set(
        ["state_backend", "uri"],
        "s3://some_bucket/some/path",
    )
    project.settings.set(["state_backend", "s3", "aws_access_key_id"], "a_different_id")
    project.settings.set(
        ["state_backend", "s3", "aws_secret_access_key"],
        "a_different_key",
    )
    s3_state_store_direct_creds: S3StateStoreManager = (
        state_store_manager_from_project_settings(project.settings)
    )
    assert s3_state_store_direct_creds.aws_access_key_id == "a_different_id"
    assert (
        s3_state_store_direct_creds.aws_secret_access_key
        == "a_different_key"  # noqa: S105
    )



from __future__ import annotations

import pytest
from flask import url_for

from meltano.api.models.security import Role, db
from meltano.api.security.identity import users


@pytest.mark.usefixtures("seed_users")
class TestRoles:
    @pytest.fixture(scope="class")
    def app(self, create_app, project):
        project.settings.config_override["ui.authentication"] = True
        return create_app()

    @pytest.mark.parametrize(
        ("user", "status_code"),
        (
            ("alice", 201),
            ("rob", 403),
            pytest.param(
                None,
                401,
                marks=pytest.mark.xfail(reason="UI/API is deprecated"),
            ),
        ),
    )
    def test_create_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            with impersonate(users.get_user(user)):
                res = api.post(
                    url_for("settings.roles"),
                    json={"role": {"name": "pytest"}},
                )

            assert res.status_code == status_code, res.data
            if status_code == 201:
                assert db.session.query(Role).filter_by(name="pytest").one()

    @pytest.mark.parametrize(
        ("user", "status_code"),
        (
            ("alice", 201),
            ("rob", 403),
            pytest.param(
                None,
                401,
                marks=pytest.mark.xfail(reason="UI/API is deprecated"),
            ),
        ),
    )
    def test_assign_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            empty_user = users.create_user(username="pytest")

            # save the new user
            db.session.commit()

            with impersonate(users.get_user(user)):
                res = api.post(
                    url_for("settings.roles"),
                    json={"role": {"name": "pytest"}, "user": empty_user.username},
                )

            assert res.status_code == status_code, res.data
            if status_code == 201:
                db.session.add(empty_user)
                assert "pytest" in empty_user.roles

    @pytest.mark.parametrize(
        ("user", "status_code"),
        (
            ("alice", 403),
            ("rob", 403),
            pytest.param(
                None,
                401,
                marks=pytest.mark.xfail(reason="UI/API is deprecated"),
            ),
        ),
    )
    def test_delete_admin_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context(), impersonate(users.get_user(user)):
            res = api.delete(
                url_for("settings.roles"),
                json={"role": {"name": "admin"}},
            )

        assert res.status_code == status_code, res.data

    @pytest.mark.parametrize(
        ("user", "status_code"),
        (
            ("alice", 201),
            ("rob", 403),
            pytest.param(
                None,
                401,
                marks=pytest.mark.xfail(reason="UI/API is deprecated"),
            ),
        ),
    )
    def test_delete_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            users.find_or_create_role(name="pytest")

            # save the new role
            db.session.commit()

            assert db.session.query(Role).filter_by(name="pytest").first()

            with impersonate(users.get_user(user)):
                res = api.delete(
                    url_for("settings.roles"),
                    json={"role": {"name": "pytest"}},
                )

            assert res.status_code == status_code, res.data
            if res.status_code == 201:
                assert not db.session.query(Role).filter_by(name="pytest").first()

    @pytest.mark.parametrize(
        ("user", "status_code"),
        (
            ("alice", 201),
            ("rob", 403),
            pytest.param(
                None,
                401,
                marks=pytest.mark.xfail(reason="UI/API is deprecated"),
            ),
        ),
    )
    def test_unassign_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            fake_role = users.find_or_create_role(name="pytest")
            empty_user = users.create_user(username="empty")
            users.add_role_to_user(empty_user, fake_role)

            # save the new user + role
            db.session.commit()

            # make sure the setup is valid
            assert fake_role in empty_user.roles

            with impersonate(users.get_user(user)):
                res = api.delete(
                    url_for("settings.roles"),
                    json={"role": {"name": "pytest"}, "user": empty_user.username},
                )

            assert res.status_code == status_code, res.data
            if res.status_code == 201:
                # refresh the instance
                empty_user = users.get_user(empty_user.username)
                assert fake_role not in empty_user.roles



"""Discover plugin definitions."""

from __future__ import annotations

import io
import logging
import re
import typing as t
from abc import ABCMeta, abstractmethod
from contextlib import suppress

import requests
from ruamel.yaml import YAMLError

import meltano
from meltano.core import bundle
from meltano.core.behavior.versioned import IncompatibleVersionError, Versioned
from meltano.core.discovery_file import DiscoveryFile
from meltano.core.plugin import BasePlugin, PluginDefinition, PluginRef, PluginType
from meltano.core.plugin.base import StandalonePlugin
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.factory import base_plugin_factory
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.utils import NotFound, find_named
from meltano.core.yaml import yaml

if t.TYPE_CHECKING:
    from meltano.core.project import Project

REQUEST_TIMEOUT_SECONDS = 30.0


class DiscoveryInvalidError(Exception):
    """The discovery.yml fails to be parsed."""


class DiscoveryUnavailableError(Exception):
    """The discovery.yml cannot be found or downloaded."""


# Increment this version number whenever the schema of discovery.yml is changed.
# See https://docs.meltano.com/contribute/plugins#discoveryyml-version for
# more information.
VERSION = 22


class PluginRepository(metaclass=ABCMeta):
    """A generic plugin definition repository."""

    @abstractmethod
    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> PluginDefinition:
        """Find a plugin definition.

        Args:
            plugin_type: The type of plugin to find.
            plugin_name: The name of the plugin to find.
            variant_name: The name of the variant to find.
        """
        ...  # noqa: WPS428

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The type of plugin to get the base plugin for.
            plugin_name: The name of the plugin to get the base plugin for.
            variant: The variant of the plugin to get the base plugin for.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
        )

        return base_plugin_factory(plugin, variant)

    def get_base_plugin(
        self,
        project_plugin: ProjectPlugin,
        **kwargs,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            project_plugin: The project plugin to get the base plugin for.
            kwargs: Additional arguments to pass to the finder.

        Returns:
            The base plugin.
        """
        plugin = project_plugin.custom_definition or self.find_definition(
            project_plugin.type,
            project_plugin.inherit_from or project_plugin.name,
            **kwargs,
        )

        return base_plugin_factory(plugin, project_plugin.variant)


class PluginDiscoveryService(  # noqa: WPS214 (too many public methods)
    PluginRepository,
    Versioned,
):
    """Discover plugin definitions."""

    __version__ = VERSION

    def __init__(self, project: Project, discovery: dict | None = None):
        """Create a new PluginDiscoveryService.

        Args:
            project: The project to discover plugins for.
            discovery: The discovery file to use.
        """
        self.project = project

        self._discovery_version = None
        self._discovery = None
        if discovery:
            self._discovery_version = DiscoveryFile.file_version(discovery)
            self._discovery = DiscoveryFile.parse(discovery)

    @property
    def file_version(self):
        """Return the version of the discovery file.

        Returns:
            The version of the discovery file.
        """
        return self._discovery_version

    @property
    def discovery_url(self):
        """Return the URL of the discovery file.

        Returns:
            The URL of the discovery file.
        """
        discovery_url = self.project.settings.get("discovery_url")

        if not discovery_url or not re.match("^https?://", discovery_url):
            return None

        return discovery_url

    @property
    def discovery_url_auth(self):
        """Return the `discovery_url_auth` setting.

        Returns:
            The `discovery_url_auth` setting.
        """
        return self.project.settings.get("discovery_url_auth")

    @property
    def discovery(self):
        """Return first compatible discovery manifest from a few locations.

        Locations:

        - project local `discovery.yml`
        - `discovery_url` project setting
        - .meltano/cache/discovery.yml
        - meltano.core.bundle

        Returns:
            The discovery file.

        Raises:
            DiscoveryInvalidError: If the discovery file is invalid.
        """
        if self._discovery:
            return self._discovery

        loaders = [
            (
                self.load_local_discovery,
                "your project's local `discovery.yml` manifest",
            ),
            (
                self.load_remote_discovery,
                f"the `discovery.yml` manifest received from {self.discovery_url}",
            ),
            (self.load_cached_discovery, "the cached `discovery.yml` manifest"),
            (self.load_bundled_discovery, "the bundled `discovery.yml` manifest"),
        ]

        errored = False
        for loader, description in loaders:
            if errored:
                logging.warning(f"Falling back on {description}...")

            try:
                loader()
            except IncompatibleVersionError as err:
                errored = True

                logging.warning(
                    f"{description.capitalize()} has version "
                    f"{err.file_version}, while this version of Meltano "
                    f"requires version {err.version}.",
                )
                if err.file_version > err.version:
                    logging.warning(
                        "Please install the latest compatible version of "
                        "Meltano using `meltano upgrade`.",
                    )
            except DiscoveryInvalidError as err:
                errored = True
                logging.error(f"{description.capitalize()} could not be parsed.")
                logging.debug(str(err))

            if self._discovery:
                return self._discovery

        raise DiscoveryInvalidError("No valid `discovery.yml` manifest could be found")

    def load_local_discovery(self) -> DiscoveryFile | None:
        """Load the local `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        with suppress(FileNotFoundError), self.project.root_dir(
            "discovery.yml",
        ).open() as local_discovery:
            return self.load_discovery(local_discovery)

    def load_remote_discovery(self) -> DiscoveryFile | None:
        """Load the remote `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        discovery_url = self.discovery_url
        if not discovery_url:
            return None

        headers = {"User-Agent": f"Meltano/{meltano.__version__}"}  # noqa: WPS609
        params = {}

        if self.discovery_url_auth:
            headers["Authorization"] = self.discovery_url_auth

        if self.project.settings.get("send_anonymous_usage_stats"):
            project_id = self.project.settings.get("project_id")

            headers["X-Project-ID"] = project_id
            params["project_id"] = project_id

        try:
            response = requests.get(
                discovery_url,
                headers=headers,
                params=params,
                timeout=REQUEST_TIMEOUT_SECONDS,
            )
            response.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as err:
            logging.debug("Remote `discovery.yml` manifest could not be downloaded.")
            logging.debug(str(err))
            return None

        remote_discovery = io.StringIO(response.text)

        return self.load_discovery(remote_discovery, cache=True)

    def load_cached_discovery(self) -> DiscoveryFile | None:
        """Load the cached `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        with suppress(
            FileNotFoundError,
        ), self.cached_discovery_file.open() as cached_discovery:
            return self.load_discovery(cached_discovery)

    def load_bundled_discovery(self):
        """Load the bundled `discovery.yml` manifest.

        Returns:
            The discovery file.
        """
        with open(bundle.root / "discovery.yml") as bundled_discovery:
            discovery = self.load_discovery(bundled_discovery, cache=True)

        return discovery

    def load_discovery(self, discovery_file, cache=False) -> DiscoveryFile:
        """Load the `discovery.yml` manifest.

        Args:
            discovery_file: The file to load.
            cache: Whether to cache the manifest.

        Returns:
            The discovery file.

        Raises:
            DiscoveryInvalidError: If the discovery file is invalid.
        """
        try:
            discovery_yaml = yaml.load(discovery_file)

            self._discovery_version = DiscoveryFile.file_version(discovery_yaml)
            self.ensure_compatible()

            self._discovery = DiscoveryFile.parse(discovery_yaml)

            if cache:
                self.cache_discovery()

            return self._discovery
        except (YAMLError, Exception) as err:
            raise DiscoveryInvalidError(str(err)) from err

    def cache_discovery(self):
        """Cache the `discovery.yml` manifest."""
        with self.cached_discovery_file.open("w") as cached_discovery:
            yaml.dump(
                self._discovery,
                cached_discovery,
            )

    @property
    def cached_discovery_file(self):
        """Return the cached `discovery.yml` manifest file.

        Returns:
            The cached `discovery.yml` manifest file.
        """
        return self.project.meltano_dir("cache", "discovery.yml")

    def get_plugins_of_type(self, plugin_type):
        """Return the plugins of the given type.

        Args:
            plugin_type: The plugin type.

        Returns:
            The list of plugins of the given type.
        """
        return self.discovery[plugin_type]

    def plugins_by_type(self):
        """Return a mapping of plugins by type.

        Returns:
            The plugins by type.
        """
        return {
            plugin_type: self.get_plugins_of_type(plugin_type)
            for plugin_type in PluginType
        }

    def plugins(self) -> t.Iterable[PluginDefinition]:
        """Generate all plugins.

        Yields:
            Each discoverable plugin.
        """
        yield from (
            plugin
            for plugin_type, plugins in self.plugins_by_type().items()
            for plugin in plugins
        )

    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,  # noqa: ARG002
    ) -> PluginDefinition:
        """Find a plugin definition by type and name.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin could not be found.
        """
        try:
            return find_named(self.get_plugins_of_type(plugin_type), plugin_name)
        except NotFound as err:
            raise PluginNotFoundError(PluginRef(plugin_type, plugin_name)) from err

    def find_definition_by_namespace(
        self,
        plugin_type: PluginType,
        namespace: str,
    ) -> PluginDefinition:
        """Find a plugin definition by type and namespace.

        Args:
            plugin_type: The plugin type.
            namespace: The plugin namespace.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin could not be found.
        """
        try:
            return next(
                plugin
                for plugin in self.get_plugins_of_type(plugin_type)
                if plugin.namespace == namespace
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def find_related_plugin_refs(
        self,
        target_plugin: ProjectPlugin,
        plugin_types: list[PluginType] | None = None,
    ):
        """Find related plugin references.

        Args:
            target_plugin: The target plugin.
            plugin_types: Types of related plugins to add.

        Returns:
            The related plugin references.
        """
        plugin_types = plugin_types or list(PluginType)

        with suppress(ValueError):
            plugin_types.remove(target_plugin.type)

        return [
            related_plugin_def
            for plugin_type in plugin_types
            for related_plugin_def in self.get_plugins_of_type(plugin_type)
            if related_plugin_def.namespace == target_plugin.namespace
        ]


class LockedDefinitionService(PluginRepository):
    """PluginRepository implementation for local files."""

    def __init__(self, project: Project) -> None:
        """Initialize the service.

        Args:
            project: The Meltano project.
        """
        self.project = project

    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> PluginDefinition:
        """Find a locked plugin definition.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant_name: The plugin variant name.

        Returns:
            The plugin definition.

        Raises:
            PluginNotFoundError: If the plugin definition could not be found.
        """
        path = self.project.plugin_lock_path(plugin_type, plugin_name, variant_name)
        try:
            standalone = StandalonePlugin.parse_json_file(path)
        except FileNotFoundError as err:
            raise PluginNotFoundError(PluginRef(plugin_type, plugin_name)) from err
        return PluginDefinition.from_standalone(standalone)

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The plugin type.
            plugin_name: The plugin name.
            variant: The plugin variant.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
            variant_name=variant,
        )

        return base_plugin_factory(plugin, plugin.variants[0])



from __future__ import annotations

import io
import json
from contextlib import contextmanager

import mock
import pytest
import requests
import requests_mock

from meltano.core import bundle
from meltano.core.plugin import PluginType, Variant, VariantNotFoundError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_discovery_service import (
    VERSION,
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.project_plugins_service import PluginAlreadyAddedException
from meltano.core.yaml import yaml

HTTP_STATUS_TEAPOT = 418


@pytest.fixture(scope="class")
def project(project):
    project.root_dir("discovery.yml").unlink()
    return project


@pytest.fixture()
def subject(plugin_discovery_service: PluginDiscoveryService):
    yield plugin_discovery_service
    plugin_discovery_service.project.settings.reset()


@pytest.fixture()
def discovery_url_mock(subject):
    with requests_mock.Mocker() as mocker:
        mocker.get(subject.discovery_url, status_code=HTTP_STATUS_TEAPOT)
        yield


@pytest.fixture(scope="class")
def tap_covid_19(project_add_service):
    try:
        plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            "tap-covid-19",
            namespace="tap-covid_19",
            pip_url="tap-covid-19",
            executable="tap-covid-19",
        )
        return project_add_service.add_plugin(plugin)
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.mark.usefixtures("discovery_url_mock")
class TestPluginDiscoveryService:
    @pytest.mark.order(0)
    @pytest.mark.meta()
    def test_discovery_url_mock(self, subject):
        assert (
            requests.get(subject.discovery_url, timeout=30.0).status_code
            == HTTP_STATUS_TEAPOT
        )

    @pytest.fixture()
    def discovery_yaml(self, subject):
        """Disable the discovery mock."""
        with subject.project.root_dir("discovery.yml").open("w") as discovery_yaml:
            yaml.dump(subject._discovery, discovery_yaml)

        subject._discovery = None

    @pytest.mark.order(1)
    def test_plugins(self, subject):
        plugins = list(subject.plugins())

        assert subject.discovery
        assert len(plugins) >= 6

    @pytest.mark.order(2)
    def test_plugins_unknown(self, subject):
        plugins = list(subject.plugins())
        assert len(plugins) >= 6

    @pytest.mark.order(3)
    def test_definition(self, subject):
        with pytest.raises(PluginNotFoundError):
            subject.find_definition(PluginType.EXTRACTORS, "unknown")

        plugin_def = subject.find_definition(PluginType.EXTRACTORS, "tap-mock")
        assert plugin_def.type == PluginType.EXTRACTORS
        assert plugin_def.name == "tap-mock"

    @pytest.mark.order(4)
    def test_find_base_plugin(self, subject):
        # If no variant is specified,
        # defaults to the first variant
        base_plugin = subject.find_base_plugin(PluginType.EXTRACTORS, "tap-mock")
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin._variant == base_plugin.variants[0]
        assert base_plugin.variant == base_plugin.variants[0].name

        base_plugin = subject.find_base_plugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant="singer-io",
        )
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "singer-io"

        base_plugin = subject.find_base_plugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant="meltano",
        )
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "meltano"

        base_plugin = subject.find_base_plugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant=Variant.ORIGINAL_NAME,
        )
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "singer-io"

        with pytest.raises(VariantNotFoundError):
            base_plugin = subject.find_base_plugin(
                PluginType.EXTRACTORS,
                "tap-mock",
                variant="unknown",
            )

    @pytest.mark.order(5)
    def test_get_base_plugin(self, subject):
        # If no variant is set on the project plugin,
        # defaults to the original variant
        project_plugin = ProjectPlugin(PluginType.EXTRACTORS, "tap-mock")
        base_plugin = subject.get_base_plugin(project_plugin)
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "singer-io"
        assert base_plugin._variant.original

        # First variant
        project_plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant="meltano",
        )
        base_plugin = subject.get_base_plugin(project_plugin)
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "meltano"

        # Another variant
        project_plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant="singer-io",
        )
        base_plugin = subject.get_base_plugin(project_plugin)
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "singer-io"

        # Original variant
        project_plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant=Variant.ORIGINAL_NAME,
        )
        base_plugin = subject.get_base_plugin(project_plugin)
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "singer-io"

        # Unknown variant
        project_plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant="unknown",
        )
        with pytest.raises(VariantNotFoundError):
            subject.get_base_plugin(project_plugin)

        # Inherited
        project_plugin = ProjectPlugin(
            PluginType.EXTRACTORS,
            "tap-mock-inherited",
            inherit_from="tap-mock",
            variant="meltano",
        )
        base_plugin = subject.get_base_plugin(project_plugin)
        assert base_plugin.type == PluginType.EXTRACTORS
        assert base_plugin.name == "tap-mock"
        assert base_plugin.variant == "meltano"

    @pytest.mark.order(6)
    @pytest.mark.usefixtures("discovery_yaml")
    def test_discovery_yaml(self, subject):
        plugins_by_type = subject.plugins_by_type()

        # raw yaml load
        for plugin_type, raw_plugin_defs in subject._discovery:
            if not PluginType.value_exists(plugin_type):
                continue

            plugin_type = PluginType(plugin_type)

            plugin_names = [plugin.name for plugin in plugins_by_type[plugin_type]]

            for raw_plugin_def in raw_plugin_defs:
                assert raw_plugin_def["name"] in plugin_names


class TestPluginDiscoveryServiceDiscoveryManifest:
    def build_discovery_yaml(self, namespace, version=VERSION):
        return {
            "version": version,
            "extractors": [{"name": f"{namespace}-test", "namespace": namespace}],
        }

    def assert_discovery_yaml(self, subject, discovery_yaml):
        subject._discovery = None
        assert (
            subject.discovery.extractors[0].namespace
            == discovery_yaml["extractors"][0]["namespace"]
        )

    @contextmanager
    def use_local_discovery(self, discovery_yaml, subject):
        local_discovery_path = subject.project.root_dir("discovery.yml")
        with local_discovery_path.open("w") as local_discovery:
            yaml.dump(discovery_yaml, local_discovery)

        yield discovery_yaml

        local_discovery_path.unlink()

    @contextmanager
    def use_remote_discovery(self, discovery_yaml, subject):
        with requests_mock.Mocker() as mocker:
            buf = io.StringIO()
            yaml.dump(discovery_yaml, buf)
            buf.seek(0)
            mocker.get(subject.discovery_url, text=buf.read())

            yield discovery_yaml

    @contextmanager
    def use_cached_discovery(self, discovery_yaml, subject):
        with subject.cached_discovery_file.open("w") as cached_discovery:
            yaml.dump(discovery_yaml, cached_discovery)

        yield discovery_yaml

        subject.cached_discovery_file.unlink()

    @pytest.fixture()
    def local_discovery(self, subject):
        with self.use_local_discovery(
            self.build_discovery_yaml("local"),
            subject,
        ) as discovery_yaml:
            yield discovery_yaml

    @pytest.fixture()
    def incompatible_local_discovery(self, subject):
        with self.use_local_discovery(
            self.build_discovery_yaml("local", version=VERSION - 1),
            subject,
        ) as discovery_yaml:
            yield discovery_yaml

    @pytest.fixture()
    def remote_discovery(
        self,
        project,  # noqa: ARG002
        subject,
    ):
        with self.use_remote_discovery(
            self.build_discovery_yaml("remote"),
            subject,
        ) as discovery_yaml:
            yield discovery_yaml

    @pytest.fixture()
    def incompatible_remote_discovery(self, subject):
        with self.use_remote_discovery(
            self.build_discovery_yaml("remote", version=VERSION + 1),
            subject,
        ) as discovery_yaml:
            yield discovery_yaml

    @pytest.fixture()
    def disabled_remote_discovery(self, subject):
        subject.project.settings.set("discovery_url", "false")

    @pytest.fixture()
    def enabled_remote_discovery_auth(self, subject):
        subject.project.settings.set("discovery_url_auth", "test")

    @pytest.fixture()
    def cached_discovery(self, subject):
        with self.use_cached_discovery(
            self.build_discovery_yaml("cached"),
            subject,
        ) as discovery_yaml:
            yield discovery_yaml

    @pytest.fixture()
    def invalid_cached_discovery(self, subject):
        with self.use_cached_discovery(
            {"version": VERSION, "invalid_key": "value"},
            subject,
        ) as discovery_yaml:
            yield discovery_yaml

    @pytest.fixture()
    def bundled_discovery(self):
        with open(bundle.root / "discovery.yml") as bundled_discovery:
            return yaml.load(bundled_discovery)

    @pytest.mark.order(7)
    def test_local_discovery(self, subject, local_discovery):
        self.assert_discovery_yaml(subject, local_discovery)

        assert not subject.cached_discovery_file.exists()

    @pytest.mark.order(8)
    @pytest.mark.usefixtures("incompatible_local_discovery")
    def test_incompatible_local_discovery(
        self,
        subject,
        remote_discovery,
    ):
        self.assert_discovery_yaml(subject, remote_discovery)

    @pytest.mark.order(9)
    def test_remote_discovery(self, subject, remote_discovery):
        self.assert_discovery_yaml(subject, remote_discovery)

        assert subject.cached_discovery_file.exists()
        with subject.cached_discovery_file.open() as cached_discovery:
            cached_discovery_yaml = yaml.load(cached_discovery)
            assert cached_discovery_yaml["version"] == remote_discovery["version"]

    @pytest.mark.order(10)
    @pytest.mark.usefixtures("enabled_remote_discovery_auth")
    @mock.patch("meltano.core.plugin_discovery_service.requests.get")
    def test_remote_discovery_with_valid_auth(
        self,
        mock_discovery_request,
        subject,
        remote_discovery,
    ):
        mock_discovery_request.return_value.status_code = 200
        mock_discovery_request.return_value.text = json.dumps(remote_discovery)

        discovery = subject.load_remote_discovery()

        mock_discovery_request.assert_called_once()
        expected_auth = subject.project.settings.get("discovery_url_auth")
        actual_auth = mock_discovery_request.call_args[1]["headers"].get(
            "Authorization",
        )
        assert expected_auth == actual_auth

        self.assert_discovery_yaml(subject, discovery)

    @pytest.mark.order(11)
    @pytest.mark.usefixtures("enabled_remote_discovery_auth")
    @mock.patch("meltano.core.plugin_discovery_service.requests.get")
    def test_remote_discovery_with_invalid_auth(
        self,
        mock_discovery_request,
        subject,
    ):
        mock_discovery_request.return_value.status_code = 401
        mock_discovery_request.return_value.raise_for_status.side_effect = (
            requests.HTTPError()
        )

        discovery = subject.load_remote_discovery()

        mock_discovery_request.assert_called_once()
        expected_auth = subject.project.settings.get("discovery_url_auth")
        actual_auth = mock_discovery_request.call_args[1]["headers"].get(
            "Authorization",
        )
        assert expected_auth == actual_auth
        assert discovery is None

    @pytest.mark.order(12)
    @mock.patch("meltano.core.plugin_discovery_service.requests.get")
    def test_remote_discovery_with_no_auth(self, mock_discovery_request, subject):
        mock_discovery_request.return_value.status_code = 401
        mock_discovery_request.return_value.raise_for_status.side_effect = (
            requests.HTTPError()
        )

        discovery = subject.load_remote_discovery()

        mock_discovery_request.assert_called_once()
        actual_auth = mock_discovery_request.call_args[1]["headers"].get(
            "Authorization",
        )
        assert actual_auth is None
        assert discovery is None

    @pytest.mark.order(13)
    @pytest.mark.usefixtures("incompatible_remote_discovery")
    def test_incompatible_remote_discovery(
        self,
        subject,
        cached_discovery,
    ):
        self.assert_discovery_yaml(subject, cached_discovery)

    @pytest.mark.order(14)
    @pytest.mark.usefixtures("disabled_remote_discovery")
    def test_disabled_remote_discovery(
        self,
        subject,
        cached_discovery,
    ):
        self.assert_discovery_yaml(subject, cached_discovery)

    @pytest.mark.order(15)
    @pytest.mark.usefixtures("incompatible_remote_discovery")
    def test_cached_discovery(
        self,
        subject,
        cached_discovery,
    ):
        self.assert_discovery_yaml(subject, cached_discovery)

    @pytest.mark.order(16)
    @pytest.mark.usefixtures(
        "incompatible_remote_discovery",
        "invalid_cached_discovery",
    )
    def test_invalid_cached_discovery(
        self,
        subject,
        bundled_discovery,
    ):
        self.assert_discovery_yaml(subject, bundled_discovery)

    @pytest.mark.order(17)
    @pytest.mark.usefixtures("incompatible_remote_discovery")
    def test_bundled_discovery(
        self,
        subject,
        bundled_discovery,
    ):
        self.assert_discovery_yaml(subject, bundled_discovery)

        assert subject.cached_discovery_file.exists()



from __future__ import annotations

import pytest
from flask import appcontext_pushed, g  # noqa: WPS347
from flask_principal import Identity, Need

from meltano.api.security.auth import ResourcePermission


@pytest.fixture()
def identity(app):
    identity = Identity("test")

    def handler(_):
        g.identity = identity

    with appcontext_pushed.connected_to(handler, app):
        yield identity


class TestResourcePermission:
    def test_allows(self, identity):
        resource_need = Need("view:design", "*")
        identity.provides.add(resource_need)

        item_need = Need("view:design", "finance")
        assert ResourcePermission(item_need).allows(identity)

    def test_allows_reject(self, identity):
        item_need = Need("view:design", "finance")
        assert not ResourcePermission(item_need).allows(identity)



from __future__ import annotations

import json as _json

import pytest
from flask import json


class TestJSON:
    @pytest.mark.parametrize(
        ("scheme", "payload", "expected"),
        (
            (
                "camel",
                {
                    "a_dict": {"camel_case": 1},
                    "a_list": [{"camel_case": 1}],
                    "a_val": "camel_value",
                    "_a_val": "camel_value",
                },
                {
                    "aDict": {"camelCase": 1},
                    "aList": [{"camelCase": 1}],
                    "aVal": "camel_value",
                    "_a_val": "camel_value",
                },
            ),
            (
                "snake",
                {
                    "aDict": {"snakeCase": 1},
                    "aList": [{"snakeCase": 1}],
                    "aVal": "snakeValue",
                    "_aVal": "snakeValue",
                },
                {
                    "a_dict": {"snake_case": 1},
                    "a_list": [{"snake_case": 1}],
                    "a_val": "snakeValue",
                    "_aVal": "snakeValue",
                },
            ),
            (
                None,
                {
                    "aDict": {"none_case": 1},
                    "a_list": [{"none_case": 1}],
                    "a_val-test": "none_value",
                },
                {
                    "aDict": {"none_case": 1},
                    "a_list": [{"none_case": 1}],
                    "a_val-test": "none_value",
                },
            ),
        ),
    )
    def test_json_scheme_encoder(self, app, scheme, payload, expected):
        with app.test_request_context(
            headers={app.config["JSON_SCHEME_HEADER"]: scheme},
        ):
            encoded = json.dumps(payload)

            # load with a Vanilla decoder
            decoded = _json.loads(encoded)

            assert decoded == expected

    @pytest.mark.parametrize(
        ("payload", "expected"),
        (
            (
                {
                    "aDict": {"camelCase": 1},
                    "aList": [{"camelCase": 1}],
                    "aVal": "simpleValue",
                    # `_` prefixed are not converted
                    "_aVal": "simpleValue",
                },
                {
                    "a_dict": {"camel_case": 1},
                    "a_list": [{"camel_case": 1}],
                    "a_val": "simpleValue",
                    # `_` prefixed are not converted
                    "_aVal": "simpleValue",
                },
            ),
        ),
    )
    def test_json_scheme_decoder(self, app, payload, expected):
        with app.test_request_context():
            encoded = _json.dumps(payload)
            decoded = json.loads(encoded)

            assert decoded == expected

    def test_json_scheme_encoder_fails(self, app):
        payload = {"snakeCase": 1, "snake_case": 2}

        with app.test_request_context():
            encoded = _json.dumps(payload)

            # if there is an issue with the encoding
            # it fallbacks to the original notation
            assert _json.loads(encoded) == payload



from __future__ import annotations

import asyncio
import json
import platform

import mock
import pytest
from click.testing import CliRunner
from mock import AsyncMock, Mock, patch

from meltano.cli import cli
from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.singer import SingerTap
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService


@pytest.fixture(scope="class")
def project_tap_mock(project_add_service):
    return project_add_service.add(PluginType.EXTRACTORS, "tap-mock")


@pytest.mark.usefixtures("project_tap_mock")
class TestCliInvoke:
    @pytest.fixture()
    def mock_invoke(self, utility, plugin_invoker_factory):
        process_mock = Mock()
        process_mock.name = "utility-mock"
        process_mock.wait = AsyncMock(return_value=0)

        with patch(
            "meltano.core.plugin_invoker.invoker_factory",
            return_value=plugin_invoker_factory,
        ), patch.object(
            ProjectPluginsService,
            "find_plugin",
            return_value=utility,
        ), patch.object(
            asyncio,
            "create_subprocess_exec",
            return_value=process_mock,
        ) as invoke_async:
            yield invoke_async

    @pytest.fixture()
    def mock_invoke_containers(self, utility, plugin_invoker_factory):
        with patch(
            "meltano.core.plugin_invoker.invoker_factory",
            return_value=plugin_invoker_factory,
        ), patch.object(
            ProjectPluginsService,
            "find_plugin",
            return_value=utility,
        ), mock.patch(
            "aiodocker.Docker",
            autospec=True,
        ) as invoke_async:
            yield invoke_async

    def test_invoke(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(cli, ["invoke", "utility-mock"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()
        args, kwargs = mock_invoke.call_args
        assert args[0].endswith("utility-mock")
        assert isinstance(kwargs, dict)

    def test_invoke_args(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(cli, ["invoke", "utility-mock", "--help"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()
        args = mock_invoke.call_args[0]
        assert args[0].endswith("utility-mock")
        assert args[1] == "--help"

    def test_invoke_command(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(
            cli,
            ["invoke", "utility-mock:cmd"],
            env={"ENV_VAR_ARG": "arg"},
        )

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()

        args = mock_invoke.call_args[0]
        assert args[0].endswith("utility-mock")
        assert args[1:] == ("--option", "arg")

    def test_invoke_command_containerized(  # noqa: WPS210
        self,
        project,
        cli_runner,
        mock_invoke_containers,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        async def async_generator(*args, **kwargs):  # noqa: ARG001
            yield "Line 1"
            yield "Line 2"  # noqa: WPS354

        docker = mock_invoke_containers.return_value
        docker_context = mock.AsyncMock()
        docker.__aenter__.return_value = docker_context  # noqa: WPS609

        container = mock.AsyncMock()
        docker_context.containers.run.return_value = container

        container.log = mock.Mock()
        container.log.return_value = async_generator()
        container.show.return_value = {"State": {"ExitCode": 0}}

        res = cli_runner.invoke(
            cli,
            ["invoke", "--containers", "utility-mock:containerized"],
            env={"SOME_ENV": "value"},
        )

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        docker_context.containers.run.assert_called_once()
        container.log.assert_called_once()
        container.wait.assert_called_once()
        container.show.assert_called_once()
        container.delete.assert_called_once()

        args, kwargs = docker_context.containers.run.call_args
        container_config = args[0]
        assert container_config["Cmd"] is None
        assert container_config["Image"] == "mock-utils/mock"
        assert "SOME_ENV=value" in container_config["Env"]

        port_bindings = container_config["HostConfig"]["PortBindings"]
        assert port_bindings == {
            "5000": [{"HostPort": "5000", "HostIP": "0.0.0.0"}],  # noqa: S104
        }

        exposed_ports = container_config["ExposedPorts"]
        assert exposed_ports == {"5000": {}}

        # Check volume env var expansion
        volume_binds = container_config["HostConfig"]["Binds"]
        assert str(project.root) in volume_binds[0]

        assert kwargs["name"].startswith("meltano-utility-mock--containerized")

        # Check env var expansion in volumes
        volume_bindings = args[0]["HostConfig"]["Binds"]
        assert volume_bindings[0].startswith(str(project.root))

    def test_invoke_command_args(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(
            cli,
            ["invoke", "utility-mock:cmd", "--verbose"],
            env={"ENV_VAR_ARG": "arg"},
        )

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()

        args = mock_invoke.call_args[0]
        assert args[0].endswith("utility-mock")
        assert args[1:] == ("--option", "arg", "--verbose")

    def test_invoke_exit_code(self, cli_runner, tap, plugin_invoker_factory, utility):
        process_mock = Mock()
        process_mock.name = "utility-mock"
        process_mock.wait = AsyncMock(return_value=2)

        with patch(
            "meltano.core.plugin_invoker.invoker_factory",
            return_value=plugin_invoker_factory,
        ), patch.object(
            ProjectPluginsService,
            "find_plugin",
            return_value=utility,
        ), patch.object(
            asyncio,
            "create_subprocess_exec",
            return_value=process_mock,
        ):
            basic = cli_runner.invoke(cli, ["invoke", tap.name])
            assert basic.exit_code == 2

    def test_invoke_triggers(
        self,
        cli_runner: CliRunner,
        project: Project,
        tap: ProjectPlugin,
    ):
        with patch.object(
            SingerTap,
            "discover_catalog",
        ) as discover_catalog, patch.object(
            SingerTap,
            "apply_catalog_rules",
        ) as apply_catalog_rules, patch.object(
            SingerTap,
            "look_up_state",
        ) as look_up_state:
            # Modes other than sync don't trigger discovery or applying catalog rules
            cli_runner.invoke(cli, ["invoke", tap.name, "--some-tap-option"])
            assert discover_catalog.call_count == 0
            assert apply_catalog_rules.call_count == 0
            assert look_up_state.call_count == 0
            project.refresh()

            # Dumping config doesn't trigger discovery or applying catalog rules
            cli_runner.invoke(cli, ["invoke", "--dump", "config", tap.name])
            assert discover_catalog.call_count == 0
            assert apply_catalog_rules.call_count == 0
            assert look_up_state.call_count == 0
            project.refresh()

            # Sync mode triggers discovery and applying catalog rules
            cli_runner.invoke(cli, ["invoke", tap.name])
            assert discover_catalog.call_count == 1
            assert apply_catalog_rules.call_count == 1
            assert look_up_state.call_count == 1
            project.refresh()

            # Dumping catalog triggers discovery and applying catalog rules
            cli_runner.invoke(cli, ["invoke", "--dump", "catalog", tap.name])
            assert discover_catalog.call_count == 2
            assert apply_catalog_rules.call_count == 2
            assert look_up_state.call_count == 2

    def test_invoke_dump_config(self, cli_runner, tap, plugin_settings_service_factory):
        settings_service = plugin_settings_service_factory(tap)

        with patch.object(SingerTap, "discover_catalog"), patch.object(
            SingerTap,
            "apply_catalog_rules",
        ):
            result = cli_runner.invoke(cli, ["invoke", "--dump", "config", tap.name])

            assert json.loads(result.stdout) == settings_service.as_dict(
                extras=False,
                process=True,
            )

    def test_list_commands(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(cli, ["invoke", "--list-commands", "utility-mock"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_not_called()
        assert "utility-mock:cmd" in res.output
        assert "description of utility command" in res.output



from __future__ import annotations

import json

from flask import Flask, url_for
from flask.testing import FlaskClient
from flask.wrappers import Response
from mock import AsyncMock, mock

from meltano.core.settings_service import REDACTED_VALUE, SettingValueStore


class TestOrchestration:
    def test_get_configuration(
        self,
        app,
        api,
        tap,
        session,
        plugin_settings_service_factory,
        monkeypatch,
    ):
        plugin_settings_service = plugin_settings_service_factory(tap)
        plugin_settings_service.set(
            "secure",
            "thisisatest",
            store=SettingValueStore.DOTENV,
            session=session,
        )

        monkeypatch.setenv("TAP_MOCK_BOOLEAN", "false")

        with app.test_request_context():
            res = api.get(
                url_for("orchestrations.get_plugin_configuration", plugin_ref=tap),
            )

            assert res.status_code == 200
            config = res.json["config"]
            config_metadata = res.json["config_metadata"]

            # make sure that set `password` is still present
            # but redacted in the response
            assert plugin_settings_service.get_with_source(
                "secure",
                session=session,
            ) == ("thisisatest", SettingValueStore.DOTENV)
            assert config["secure"] == REDACTED_VALUE
            assert config_metadata["secure"]["redacted"] is True
            assert config_metadata["secure"]["source"] == "dotenv"
            assert config_metadata["secure"]["auto_store"] == "dotenv"
            assert config_metadata["secure"]["overwritable"] is True

            # make sure the `hidden` setting is still present
            # but hidden in the response
            assert plugin_settings_service.get_with_source(
                "hidden",
                session=session,
            ) == (42, SettingValueStore.DEFAULT)
            assert "hidden" not in config

            # Extras are excluded
            assert "_select" not in config
            assert "_metadata" not in config
            assert "_schema" not in config

            setting_names = [s["name"] for s in res.json["settings"]]
            assert "_select" not in setting_names
            assert "_metadata" not in setting_names
            assert "_schema" not in setting_names

    def test_save_configuration(
        self,
        app,
        api,
        tap,
        session,
        plugin_settings_service_factory,
    ):
        plugin_settings_service = plugin_settings_service_factory(tap)
        plugin_settings_service.set(
            "secure",
            "thisisatest",
            store=SettingValueStore.DOTENV,
            session=session,
        )
        plugin_settings_service.set(
            "protected",
            "iwontchange",
            store=SettingValueStore.DB,
            session=session,
        )

        with app.test_request_context():
            res = api.put(
                url_for("orchestrations.save_plugin_configuration", plugin_ref=tap),
                json={"config": {"protected": "N33DC0FF33", "secure": "newvalue"}},
            )

            assert res.status_code == 200
            config = res.json["config"]

            # make sure that set `password` has been updated
            # but redacted in the response
            assert plugin_settings_service.get_with_source(
                "secure",
                session=session,
            ) == ("newvalue", SettingValueStore.DOTENV)
            assert config["secure"] == REDACTED_VALUE

            # make sure the `readonly` field has not been updated
            assert plugin_settings_service.get_with_source(
                "protected",
                session=session,
            ) == ("iwontchange", SettingValueStore.DB)

            # make sure the `hidden` setting is still present
            # but hidden in the response
            assert plugin_settings_service.get_with_source(
                "hidden",
                session=session,
            ) == (42, SettingValueStore.DEFAULT)
            assert "hidden" not in config

    @mock.patch("meltano.core.plugin_test_service.PluginInvoker.invoke_async")
    def test_test_plugin_configuration_success(
        self,
        mock_invoke_async,
        app: Flask,
        api: FlaskClient,
        tap,
    ):
        mock_invoke = mock.Mock()
        mock_invoke.sterr.at_eof.side_effect = True
        mock_invoke.stdout.at_eof.side_effect = (False, True)
        mock_invoke.wait = AsyncMock(return_value=-1)
        mock_invoke.returncode = -1
        payload = json.dumps({"type": "RECORD"}).encode()
        mock_invoke.stdout.readline = AsyncMock(return_value=b"%b" % payload)

        mock_invoke_async.return_value = mock_invoke

        with app.test_request_context():
            url = url_for("orchestrations.test_plugin_configuration", plugin_ref=tap)
            res: Response = api.post(url, json={})

        assert res.status_code == 200
        assert res.json["is_success"]

    @mock.patch("meltano.core.plugin_test_service.PluginInvoker.invoke_async")
    def test_test_plugin_configuration_failure(
        self,
        mock_invoke_async,
        app: Flask,
        api: FlaskClient,
        tap,
    ):
        mock_invoke = mock.Mock()
        mock_invoke.sterr.at_eof.side_effect = True
        mock_invoke.stdout.at_eof.side_effect = (False, True)
        mock_invoke.wait = AsyncMock(return_value=0)
        mock_invoke.returncode = 0
        payload = b"test"
        mock_invoke.stdout.readline = AsyncMock(return_value=b"%b" % payload)

        mock_invoke_async.return_value = mock_invoke

        with app.test_request_context():
            url = url_for("orchestrations.test_plugin_configuration", plugin_ref=tap)
            res: Response = api.post(url, json={})

        assert res.status_code == 200
        assert not res.json["is_success"]



from __future__ import annotations

import json
import logging
import subprocess
import typing as t
import uuid
from contextlib import contextmanager, suppress
from http import server as server_lib
from threading import Thread
from time import sleep

import mock
import pytest
from snowplow_tracker import Emitter

from meltano.core.project import Project
from meltano.core.tracking.contexts.cli import CliEvent
from meltano.core.tracking.contexts.environment import EnvironmentContext
from meltano.core.tracking.contexts.exception import ExceptionContext
from meltano.core.tracking.contexts.project import ProjectContext
from meltano.core.tracking.tracker import TelemetrySettings, Tracker
from meltano.core.utils import hash_sha256

if t.TYPE_CHECKING:
    from fixtures.docker import SnowplowMicro


def load_analytics_json(project: Project) -> dict[str, t.Any]:
    with open(project.meltano_dir() / "analytics.json") as analytics_json_file:
        return json.load(analytics_json_file)


def check_analytics_json(project: Project) -> None:
    analytics_json = load_analytics_json(project)

    for key in ("client_id", "project_id"):
        value = analytics_json[key]
        assert isinstance(value, str)
        uuid.UUID(value)

    assert isinstance(analytics_json["send_anonymous_usage_stats"], bool)


@contextmanager
def delete_analytics_json(project: Project) -> None:
    # After Python 3.7 support is dropped, use `.unlink(missing_ok=True)`
    # instead of a suppression.
    with suppress(FileNotFoundError):
        (project.meltano_dir() / "analytics.json").unlink()
    try:
        yield
    finally:
        Tracker(project)


class TestTracker:
    @pytest.fixture(autouse=True)
    def clear_telemetry_settings(self, project, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.delenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", raising=False)
        config = project.settings.meltano_yml_config
        config.pop("send_anonymous_usage_stats", None)
        project.settings.update_meltano_yml_config(config)
        original_config_override = (
            type(project.settings).config_override.copy(),
            project.settings.config_override.copy(),
        )
        try:
            type(project.settings).config_override.clear()
            project.settings.config_override.clear()
            project.refresh()
            yield
        finally:
            (  # noqa: WPS414
                type(project.settings).config_override,
                project.settings.config_override,
            ) = original_config_override

    def test_telemetry_state_change_check(self, project: Project):
        with mock.patch.object(
            Tracker,
            "save_telemetry_settings",
        ) as mocked, delete_analytics_json(project):
            Tracker(project)
            assert mocked.call_count == 1

    def test_update_analytics_json(self, project: Project):
        tracker = Tracker(project)
        inital_send_anonymous_usage_stats = tracker.send_anonymous_usage_stats

        # Check the inital value of `send_anonymous_usage_stats`
        analytics_json_pre = load_analytics_json(project)
        assert (
            inital_send_anonymous_usage_stats
            == analytics_json_pre["send_anonymous_usage_stats"]
        )

        # Flip the value of `send_anonymous_usage_stats`
        tracker.send_anonymous_usage_stats = not inital_send_anonymous_usage_stats
        tracker.telemetry_state_change_check(
            TelemetrySettings(
                analytics_json_pre["client_id"],
                analytics_json_pre["project_id"],
                analytics_json_pre["send_anonymous_usage_stats"],
            ),
        )

        # Ensure `send_anonymous_usage_stats` has been flipped on disk
        analytics_json_post = load_analytics_json(project)
        assert (
            inital_send_anonymous_usage_stats
            != analytics_json_post["send_anonymous_usage_stats"]
        )

    def test_restore_project_id_from_analytics_json(self, project: Project):
        Tracker(project)  # Ensure `analytics.json` exists and is valid

        original_project_id = project.settings.get("project_id")

        # Delete the project ID from `meltano.yml`,
        # but leave it unchanged in `analytics.json`
        config = project.settings.meltano_yml_config
        del config["project_id"]
        project.settings.update_meltano_yml_config(config)
        # The project refreshes itself once the config has been changed, which
        # creates a new `ProjectSettingsService`, restoring the `project_id`.
        restored_project_id = project.settings.get("project_id")

        # Depending on what tests were run before this one, the project ID
        # might have been randomly generated, or taken from `analytics.json`,
        # so we accept the restored one if it is equal to the original, or if
        # it is equal after the same transformation that gets applied to the
        # project ID when it is originally stored in `analytics.json`.
        assert original_project_id == restored_project_id or (
            str(uuid.UUID(hash_sha256(original_project_id)[::2])) == restored_project_id
        )

    def test_no_project_id_state_change_if_tracking_disabled(self, project: Project):
        method_name = "track_telemetry_state_change_event"

        project.settings.set("send_anonymous_usage_stats", True)
        project.settings.set("project_id", str(uuid.uuid4()))
        Tracker(project).save_telemetry_settings()

        project.settings.set("send_anonymous_usage_stats", False)
        with mock.patch.object(Tracker, method_name) as mocked:
            Tracker(project).save_telemetry_settings()
            assert mocked.call_count == 1

        project.settings.set("project_id", str(uuid.uuid4()))
        with mock.patch.object(Tracker, method_name) as mocked:
            Tracker(project).save_telemetry_settings()
            assert mocked.call_count == 0

    def test_no_state_change_event_without_analytics_json(self, project: Project):
        method_name = "track_telemetry_state_change_event"

        project.settings.set("send_anonymous_usage_stats", True)
        project.settings.set("project_id", str(uuid.uuid4()))
        Tracker(project).save_telemetry_settings()

        with delete_analytics_json(project):
            project.settings.set("project_id", str(uuid.uuid4()))
            with mock.patch.object(Tracker, method_name) as mocked:
                Tracker(project)
                assert mocked.call_count == 0

    def test_analytics_json_is_created(self, project: Project):
        Tracker(project)
        check_analytics_json(project)
        with delete_analytics_json(project):
            Tracker(project)
            check_analytics_json(project)

    @pytest.mark.parametrize(
        "analytics_json_content",
        (
            f'{{"clientId":"{str(uuid.uuid4())}","project_id":"{str(uuid.uuid4())}","send_anonymous_usage_stats":true}}',  # noqa: E501
            f'{{"client_id":"{str(uuid.uuid4())}","projectId":"{str(uuid.uuid4())}","send_anonymous_usage_stats":true}}',  # noqa: E501
            f'{{"client_id":"{str(uuid.uuid4())}","project_id":"{str(uuid.uuid4())}","send_anon_usage_stats":true}}',  # noqa: E501
            f'["{str(uuid.uuid4())}","{str(uuid.uuid4())}", true]',
            f'client_id":"{str(uuid.uuid4())}","project_id":"{str(uuid.uuid4())}","send_anonymous_usage_stats":true}}',  # noqa: E501
        ),
        ids=(0, 1, 2, 3, 4),
    )
    def test_invalid_analytics_json_is_overwritten(
        self,
        project: Project,
        analytics_json_content: str,
    ):
        with delete_analytics_json(project):
            # Use `delete_analytics_json` to ensure `analytics.json` is restored after
            analytics_json_path = project.meltano_dir() / "analytics.json"
            with open(analytics_json_path, "w") as analytics_json_file:
                analytics_json_file.write(analytics_json_content)

            with pytest.raises((TypeError, KeyError, json.JSONDecodeError)):
                check_analytics_json(project)

            Tracker(project)

            check_analytics_json(project)

    def test_restore_project_id_and_telemetry_state_change(self, project: Project):
        """
        Test that `project_id` is restored from `analytics.json`, and a telemetry state
        change event is fired because `send_anonymous_usage_stats` is negated.
        """  # noqa: D205, D400
        Tracker(project)  # Ensure `analytics.json` exists and is valid

        original_project_id = project.settings.get("project_id")

        # Delete project ID from `meltano.yml`; leave it unchanged in `analytics.json`
        config = project.settings.meltano_yml_config.copy()
        del config["project_id"]
        config["send_anonymous_usage_stats"] = not load_analytics_json(project)[
            "send_anonymous_usage_stats"
        ]
        project.settings.update_meltano_yml_config(config)
        # The project refreshes itself once the config has been changed, which
        # creates a new `ProjectSettingsService`, restoring the `project_id`.
        restored_project_id = project.settings.get("project_id")

        # Depending on what tests were run before this one, the project ID
        # might have been randomly generated, or taken from `analytics.json`,
        # so we accept the restored one if it is equal to the original, or if
        # it is equal after the same transformation that gets applied to the
        # project ID when it is originally stored in `analytics.json`.
        assert original_project_id == restored_project_id or (
            str(uuid.UUID(hash_sha256(original_project_id)[::2])) == restored_project_id
        )

        with mock.patch.object(Tracker, "track_telemetry_state_change_event") as mocked:
            original_config_override = project.settings.config_override.copy()
            try:
                project.settings.config_override.pop("send_anonymous_usage_stats", None)
                Tracker(project)
                assert mocked.call_count == 1
            finally:
                project.settings.config_override = original_config_override

    @pytest.mark.parametrize(
        ("snowplow_endpoints", "send_stats", "expected"),
        (
            (["https://example.com"], True, True),
            (["https://example.com"], False, False),
            ([], True, False),
            ([], False, False),
        ),
    )
    def test_can_track(
        self,
        project: Project,
        snowplow_endpoints: list[str],
        send_stats: bool,
        expected: bool,
    ):
        project.settings.set("snowplow.collector_endpoints", snowplow_endpoints)
        project.settings.set("send_anonymous_usage_stats", send_stats)
        assert Tracker(project).can_track() is expected

    def test_send_anonymous_usage_stats(self, project: Project, monkeypatch):
        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "True")
        assert Tracker(project).send_anonymous_usage_stats is True

        # Ensure the env var takes priority
        project.settings.set("send_anonymous_usage_stats", False)
        assert Tracker(project).send_anonymous_usage_stats is True

        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "False")
        assert Tracker(project).send_anonymous_usage_stats is False

        # Ensure the env var takes priority
        project.settings.set("send_anonymous_usage_stats", True)
        assert Tracker(project).send_anonymous_usage_stats is False

    @pytest.mark.parametrize("setting_value", (False, True))
    def test_send_anonymous_usage_stats_no_env(
        self,
        project: Project,
        setting_value: bool,
    ):
        project.settings.set("send_anonymous_usage_stats", setting_value)
        assert Tracker(project).send_anonymous_usage_stats is setting_value

    def test_default_send_anonymous_usage_stats(self, project: Project):
        assert Tracker(project).send_anonymous_usage_stats

    @pytest.mark.usefixtures("project")
    def test_exit_event_is_fired(self, snowplow: SnowplowMicro):
        subprocess.run(("meltano", "invoke", "alpha-beta-fox"))

        event_summary = snowplow.all()
        assert event_summary["good"] > 0
        assert event_summary["bad"] == 0

        exit_event = snowplow.good()[0]["event"]
        assert exit_event["event_name"] == "exit_event"
        assert exit_event["unstruct_event"]["data"]["data"]["exit_code"] == 1

    @pytest.mark.parametrize("send_anonymous_usage_stats", (True, False))
    def test_context_with_telemetry_state_change_event(
        self,
        project: Project,
        send_anonymous_usage_stats: bool,
    ):
        tracker = Tracker(project)
        tracker.send_anonymous_usage_stats = send_anonymous_usage_stats

        passed = False

        class MockSnowplowTracker:
            def track_unstruct_event(self, _, contexts):
                # Can't put asserts in here because this method is executed
                # withing a try-except block that catches all exceptions.
                nonlocal passed
                expected_contexts = [EnvironmentContext, ProjectContext]
                if send_anonymous_usage_stats:
                    expected_contexts.append(ExceptionContext)
                passed = len(set(contexts)) == len(expected_contexts) and all(
                    isinstance(ctx, tuple(expected_contexts)) for ctx in contexts
                )

        tracker.snowplow_tracker = MockSnowplowTracker()

        tracker.track_telemetry_state_change_event(
            "project_id",
            uuid.uuid4(),
            uuid.uuid4(),
        )
        assert passed

        tracker.track_telemetry_state_change_event(
            "send_anonymous_usage_stats",
            True,
            False,
        )
        assert passed

        tracker.track_telemetry_state_change_event(
            "send_anonymous_usage_stats",
            False,
            True,
        )
        assert passed

    @pytest.mark.parametrize(
        ("sleep_duration", "timeout_should_occur"),
        ((1.0, False), (5.0, True)),
        ids=("no_timeout", "timeout"),
    )
    def test_timeout_if_endpoint_unavailable(
        self,
        project: Project,
        sleep_duration,
        timeout_should_occur,
    ):
        """Test to ensure that the default tracker timeout is respected.

        An HTTP sever is run on another thread, which handles request sent from
        the Snowplow tracker. When `timeout_should_occur` is `False`, we check
        that this sever is properly masqerading as a Snowplow endpoint. When
        `timeout_should_occur` is `True`, we check that when the server takes
        too long to respond, we timeout and continue without raising an error.
        """
        timeout_occured = False

        class HTTPRequestHandler(server_lib.SimpleHTTPRequestHandler):
            def do_GET(self) -> None:  # noqa: N802
                sleep(sleep_duration)
                self.send_response(200, "OK")
                self.end_headers()

        server = server_lib.HTTPServer(("localhost", 0), HTTPRequestHandler)
        server_thread = Thread(
            target=server.serve_forever,
            kwargs={"poll_interval": 0.1},
        )
        server_thread.start()

        project.settings.set(
            "snowplow.collector_endpoints",
            f'["http://localhost:{server.server_port}"]',
        )

        def emitter_failure_callback(_, failure_events: list):  # noqa: ARG001
            nonlocal timeout_occured
            # `timeout_occured` is technically a misnomer here, since there are
            # multiple reasons for failure, but this callback doesn't let us
            # distinguish. We can rely on the `timeout_should_occur is True`
            # case to ensure that this callback won't be called for non-timeout
            # reasons.
            timeout_occured = True

        tracker = Tracker(project)
        assert len(tracker.snowplow_tracker.emitters) == 1
        tracker.snowplow_tracker.emitters[0].on_failure = emitter_failure_callback

        tracker.track_command_event(CliEvent.started)

        server.shutdown()
        server_thread.join()

        assert timeout_occured is timeout_should_occur

    def test_project_context_send_anonymous_usage_stats_source(
        self,
        project: Project,
        monkeypatch,
    ):
        def get_source():
            return ProjectContext(project, uuid.uuid4()).to_json()["data"][
                "send_anonymous_usage_stats_source"
            ]

        assert get_source() == "default"

        project.settings.set(
            "send_anonymous_usage_stats",
            not project.settings.get("send_anonymous_usage_stats"),
        )
        assert get_source() == "meltano_yml"

        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "True")
        assert get_source() == "env"

    def test_get_snowplow_tracker_invalid_endpoint(
        self,
        project: Project,
        caplog,
        monkeypatch,
    ):
        endpoints = """
            [
                "notvalid:8080",
                "https://valid.endpoint:8080",
                "file://bad.scheme",
                "https://other.endpoint/path/to/collector"
            ]
        """
        monkeypatch.setenv("MELTANO_SNOWPLOW_COLLECTOR_ENDPOINTS", endpoints)
        logging.getLogger().setLevel(logging.DEBUG)

        with caplog.at_level(logging.WARNING, logger="snowplow_tracker.emitters"):
            tracker = Tracker(project)

        try:
            assert caplog.records[0].levelname == "WARNING"
            assert caplog.records[0].msg["event"] == "invalid_snowplow_endpoint"
            assert caplog.records[0].msg["endpoint"] == "notvalid:8080"

            assert caplog.records[1].levelname == "WARNING"
            assert caplog.records[1].msg["event"] == "invalid_snowplow_endpoint"
            assert caplog.records[1].msg["endpoint"] == "file://bad.scheme"

            assert len(tracker.snowplow_tracker.emitters) == 2
            assert isinstance(tracker.snowplow_tracker.emitters[0], Emitter)
            assert (
                tracker.snowplow_tracker.emitters[0].endpoint
                == "https://valid.endpoint:8080/i"
            )

            assert isinstance(tracker.snowplow_tracker.emitters[1], Emitter)
            assert (
                tracker.snowplow_tracker.emitters[1].endpoint
                == "https://other.endpoint/path/to/collector/i"
            )
        finally:
            # Remove the seemingly valid emitters to prevent a logging error on exit.
            tracker.snowplow_tracker.emitters = []

    def test_client_id_from_env_var(
        self,
        project: Project,
        monkeypatch: pytest.MonkeyPatch,
    ):
        with delete_analytics_json(project):
            monkeypatch.setenv("MELTANO_CLIENT_ID", "invalid-context-uuid")
            with pytest.warns(RuntimeWarning, match="Invalid telemetry client UUID"):
                # Ensure it generated a random UUID as a fallback
                uuid.UUID(str(Tracker(project).client_id))

            ctx_id = uuid.uuid4()
            monkeypatch.setenv("MELTANO_CLIENT_ID", str(ctx_id))
            # Ensure it takes the client ID from the env var
            assert Tracker(project).client_id == ctx_id

            monkeypatch.delenv("MELTANO_CLIENT_ID")
            # Ensure it uses the client ID stored in `analytics.json`
            assert Tracker(project).client_id == ctx_id

            ctx_id_2 = uuid.uuid4()
            monkeypatch.setenv("MELTANO_CLIENT_ID", str(ctx_id_2))
            # Ensure the env var takes priority over `analytics.json`
            assert Tracker(project).client_id == ctx_id_2

            monkeypatch.delenv("MELTANO_CLIENT_ID")
            # Ensure the new client ID overwrites the old one in `analytics.json`
            assert Tracker(project).client_id == ctx_id_2



from __future__ import annotations

from datetime import datetime
from http import HTTPStatus

import mock
import pytest
from flask import url_for
from flask_login import current_user
from flask_security import AnonymousUser, login_user
from freezegun import freeze_time
from sqlalchemy.orm import joinedload

from meltano.api.models.oauth import OAuth
from meltano.api.models.security import User, db
from meltano.api.security.identity import FreeUser, users
from meltano.api.security.oauth import OAuthError, gitlab_token_identity
from meltano.core.project import PROJECT_READONLY_ENV, Project

STATUS_READONLY = 499


def gitlab_client():
    client_mock = mock.Mock()
    client_mock.auth.return_value = None
    user = mock.Mock(username="gitlabfan", email="valid@test.com", state="active", id=1)
    type(client_mock).user = mock.PropertyMock(return_value=user)
    return client_mock


class TestFreeUser:
    def test_all_roles(self):
        assert len(FreeUser().roles) == 2
        role = users.find_or_create_role("this_is_a_test")
        assert FreeUser().has_role(role)


@pytest.mark.usefixtures("seed_users")
class TestNothingEnabled:
    @pytest.fixture(scope="class")
    def app(self, create_app):
        return create_app()

    def test_current_user(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), FreeUser)

    def test_identity(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["anonymous"] is True
            assert res.json["can_sign_in"] is False

    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    def test_upgrade(self, app, api):
        with app.test_request_context():
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.CREATED
            assert res.data == b"Meltano update in progress."

    def test_plugins(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    def test_plugins_add(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == HTTPStatus.OK
            assert res.json["name"] == "tap-gitlab"


@pytest.mark.usefixtures("seed_users")
class TestProjectReadonlyEnabled:
    @pytest.fixture(scope="class")
    def project(self, project):
        Project.deactivate()
        with pytest.MonkeyPatch.context() as monkeypatch:
            monkeypatch.setenv(PROJECT_READONLY_ENV, "true")
            project.refresh()
            yield project

    def test_current_user(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), FreeUser)

    def test_identity(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["anonymous"] is True
            assert res.json["can_sign_in"] is False

    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    def test_upgrade(self, app, api):
        with app.test_request_context():
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.CREATED
            assert res.data == b"Meltano update in progress."

    def test_plugins(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    def test_plugins_add(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == STATUS_READONLY
            assert b"deployed as read-only" in res.data

    @pytest.mark.xfail(reason="UI/API is deprecated")
    @pytest.mark.usefixtures("tap", "target")
    def test_pipeline_schedules_save(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("orchestrations.save_pipeline_schedule"),
                json={
                    "name": "mock-to-mock",
                    "extractor": "tap-mock",
                    "loader": "target-mock",
                    "transform": "skip",
                    "interval": "@once",
                },
            )

            assert res.status_code == STATUS_READONLY
            assert b"deployed as read-only" in res.data


@pytest.mark.usefixtures("seed_users")
class TestReadonlyEnabled:
    @pytest.fixture(scope="class")
    def app(self, create_app, project):
        with pytest.MonkeyPatch.context() as monkeypatch:
            monkeypatch.setitem(
                type(project.settings).config_override,
                "ui.readonly",
                True,
            )
            project.refresh()
            yield create_app()

    def test_current_user(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), FreeUser)

    def test_identity(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["anonymous"] is True
            assert res.json["can_sign_in"] is False

    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    def test_upgrade(self, app, api):
        with app.test_request_context():
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == STATUS_READONLY
            assert b"read-only mode" in res.data

    def test_plugins(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    def test_plugins_add(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == STATUS_READONLY
            assert b"read-only mode" in res.data


@pytest.mark.usefixtures("seed_users")
class TestAuthenticationEnabled:
    @pytest.fixture(scope="class")
    def app(self, create_app, project):
        with pytest.MonkeyPatch.context() as monkeypatch:
            monkeypatch.setitem(
                type(project.settings).config_override,
                "ui.readonly",
                True,
            )
            project.refresh()
            yield create_app()

    @mock.patch("gitlab.Gitlab", return_value=gitlab_client())
    def test_gitlab_token_identity_creates_user(self, gitlab, app):  # noqa: ARG002
        token = {
            "access_token": "thisisavalidtoken",
            "id_token": "thisisavalidJWT",
            "created_at": 1548789020,
        }

        # test automatic user creation
        with app.test_request_context("/oauth/authorize"):
            gitlab_token_identity(token)
            assert (
                db.session.query(OAuth)
                .options(joinedload(OAuth.user))
                .filter(
                    OAuth.access_token == token["access_token"]  # noqa: WPS222
                    and OAuth.id_token == token["id_token"]
                    and OAuth.provider_user_id == user.id  # noqa: F821
                    and OAuth.provider_id == "gitlab"
                    and User.email == user.email,  # noqa: F821
                )
                .first()
            )

    @mock.patch("gitlab.Gitlab", return_value=gitlab_client())
    def test_gitlab_token_identity_maps_user(self, gitlab, app):  # noqa: ARG002
        token = {
            "access_token": "thisisavalidtoken",
            "id_token": "thisisavalidJWT",
            "created_at": 1548789020,
        }

        # test automatic user mapping
        with app.test_request_context("/oauth/authorize"):
            # let's create a user with the same email, that is currently logged
            user = users.create_user(email="valid@test.com")

            # but only if the user is currently logged (to prevent hi-jacking)
            with pytest.raises(OAuthError):
                identity = gitlab_token_identity(token)

            # the new identity should be mapped to the existing user
            login_user(user)
            identity = gitlab_token_identity(token)
            assert identity.user == user

    @freeze_time("2000-01-01")
    def test_login_audit_columns(self, app):
        with app.test_request_context():
            alice = users.get_user("alice")
            login_count = alice.login_count

            login_user(alice)

            # time is frozen, so it should work
            assert alice.last_login_at == datetime.utcnow()
            assert alice.login_count == login_count + 1

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_current_user(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), AnonymousUser)

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_identity(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    def test_identity_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["username"] == "alice"
            assert res.json["anonymous"] is False
            assert res.json["can_sign_in"] is False

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location.startswith(url_for("security.login"))

    def test_bootstrap_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_upgrade(self, app, api):
        with app.test_request_context():
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_upgrade_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.CREATED
            assert res.data == b"Meltano update in progress."

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_plugins(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    def test_plugins_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_plugins_add(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    @pytest.mark.xfail(reason="UI/API is deprecated")
    def test_plugins_add_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == HTTPStatus.OK
            assert res.json["name"] == "tap-gitlab"


@pytest.mark.usefixtures("seed_users")
class TestAuthenticationAndReadonlyEnabled:
    @pytest.fixture(scope="class")
    def app(self, create_app, project):
        with pytest.MonkeyPatch.context() as monkeypatch:
            for setting in ("ui.readonly", "ui.authentication"):
                monkeypatch.setitem(
                    type(project.settings).config_override,
                    setting,
                    True,
                )
            project.refresh()
            yield create_app()

    def test_current_user(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), AnonymousUser)

    def test_identity(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    def test_identity_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["username"] == "alice"
            assert res.json["anonymous"] is False
            assert res.json["can_sign_in"] is False

    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location.startswith(url_for("security.login"))

    def test_bootstrap_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    def test_upgrade(self, app, api):
        with app.test_request_context():
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    def test_upgrade_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == STATUS_READONLY
            assert b"read-only mode" in res.data

    def test_plugins(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    def test_plugins_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    def test_plugins_add(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == HTTPStatus.UNAUTHORIZED
            assert res.data == b"Authentication is required to access this resource."

    def test_plugins_add_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == STATUS_READONLY
            assert b"read-only mode" in res.data


@pytest.mark.usefixtures("seed_users")
class TestAuthenticationAndAnonymousReadonlyEnabled:
    @pytest.fixture(scope="class")
    def app(self, create_app, project):
        with pytest.MonkeyPatch.context() as monkeypatch:
            for setting in ("ui.authentication", "ui.anonymous_readonly"):
                monkeypatch.setitem(
                    type(project.settings).config_override,
                    setting,
                    True,
                )
            project.refresh()
            yield create_app()

    def test_current_user(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), AnonymousUser)

    def test_identity(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["anonymous"] is True
            assert res.json["can_sign_in"] is True

    def test_identity_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("api_root.identity"))

            assert res.status_code == HTTPStatus.OK
            assert res.json["username"] == "alice"
            assert res.json["anonymous"] is False
            assert res.json["can_sign_in"] is False

    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    def test_bootstrap_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("root.bootstrap"))

            assert res.status_code == HTTPStatus.FOUND
            assert res.location == url_for("root.default")

    def test_upgrade(self, app, api):
        with app.test_request_context():
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.FORBIDDEN
            assert res.data == b"You do not have the required permissions."

    def test_upgrade_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.post(url_for("api_root.upgrade"))

            assert res.status_code == HTTPStatus.CREATED
            assert res.data == b"Meltano update in progress."

    def test_plugins(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    def test_plugins_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.get(url_for("plugins.all"))

            assert res.status_code == HTTPStatus.OK
            assert "extractors" in res.json

    def test_plugins_add(self, app, api):
        with app.test_request_context():
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == STATUS_READONLY
            assert b"read-only mode until you sign in" in res.data

    def test_plugins_add_authenticated(self, app, api, impersonate):
        with app.test_request_context(), impersonate(users.get_user("alice")):
            res = api.post(
                url_for("plugins.add"),
                json={"plugin_type": "extractors", "name": "tap-gitlab"},
            )

            assert res.status_code == HTTPStatus.OK
            assert res.json["name"] == "tap-gitlab"



from __future__ import annotations

import pytest

from meltano.api.models import db as _db


@pytest.fixture(autouse=True)
def session(app, connection):
    # attach it to our test app
    _db.app = app

    # create a new session upon our auto-rollback
    # connection
    _db.session = _db.create_scoped_session(options={"bind": connection, "binds": {}})

    try:
        yield _db.session
    finally:
        _db.session.remove()



from __future__ import annotations

import json

import pytest

from meltano.api.mail import MailService
from meltano.api.models.subscription import Subscription, SubscriptionEventType
from meltano.core.project_settings_service import ProjectSettingsService


class TestMailService:
    def test_get_unsubscribe_group(self, project):
        default_id = 12751
        custom_id = 42

        sub = Subscription()
        sub.event_type = SubscriptionEventType.PIPELINE_MANUAL_RUN

        settings = ProjectSettingsService(project)
        assert settings.get("mail.sendgrid_unsubscribe_group_id") == default_id

        # the default case
        service = MailService(project)

        assert service.get_unsubscribe_group(sub) == default_id

        # the custom case
        settings = ProjectSettingsService(project)
        settings.set("mail.sendgrid_unsubscribe_group_id", custom_id)
        assert settings.get("mail.sendgrid_unsubscribe_group_id") == custom_id

        custom_service = MailService(project)
        assert custom_service.get_unsubscribe_group(sub) == custom_id

    @pytest.mark.order(after="test_get_unsubscribe_group")
    def test_create_message(self, project):
        sub = Subscription()
        sub.event_type = SubscriptionEventType.PIPELINE_MANUAL_RUN
        sub.id = 42
        sub.recipient = "test@test.com"

        settings = ProjectSettingsService(project)
        settings.set("mail.sendgrid_unsubscribe_group_id", 1)

        service = MailService(project)
        message = service.create_message(sub)

        assert "test@test.com" in message.recipients

        decoded_headers = json.loads(message.extra_headers["X-SMTPAPI"])
        assert decoded_headers.get("asm_group_id") == 1
        assert decoded_headers.get("unique_args").get("subscription_id") == str(sub.id)



from __future__ import annotations

import logging
from contextlib import contextmanager

import pytest
from flask import request_started
from flask_security.utils import login_user, logout_user

from meltano.api import app as meltano_app
from meltano.api.security.identity import create_dev_user
from meltano.core.project import Project


@pytest.fixture()
def impersonate(
    app,  # noqa: ARG001
):
    @contextmanager
    def factory(user):
        def push(_):
            if user:
                login_user(user)
            else:
                logout_user()

        with request_started.connected_to(push):
            yield

    return factory


@pytest.fixture(scope="class")
def app(create_app):
    root_logger = logging.getLogger()
    log_level = root_logger.level
    try:
        yield create_app()
    finally:
        root_logger.setLevel(log_level)


@pytest.fixture(scope="class")
def create_app(
    request,
    project,  # noqa: ARG001
):
    def _factory(**kwargs):
        config = {"TESTING": True, "LOGIN_DISABLED": False, "ENV": "test", **kwargs}

        app = meltano_app.create_app(config)

        # let's push an application context so the
        # `current_app` is ready in each test
        ctx = app.app_context()
        ctx.push()

        # let's make sure to pop the context at the end
        request.addfinalizer(lambda: ctx.pop())

        return app

    return _factory


@pytest.fixture()
def api(app, project: Project):
    project.refresh()
    return app.test_client()


@pytest.fixture()
def seed_users(
    app,  # noqa: ARG001
    session,  # noqa: ARG001
) -> None:
    create_dev_user()



from __future__ import annotations

from collections import Counter

import mock

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.plugin.base import PluginType
from meltano.core.project import Project


class TestCliDiscover:
    def test_discover(
        self,
        project: Project,
        cli_runner,
        hub_request_counter: Counter,
    ):
        adapter = project.hub_service.session.get_adapter(
            project.hub_service.hub_api_url,
        )

        with mock.patch("requests.adapters.HTTPAdapter.send", adapter.send):
            result = cli_runner.invoke(cli, ["discover"])

        assert_cli_runner(result)

        for plugin_type in PluginType:
            request_count = 1 if plugin_type.discoverable else 0
            assert hub_request_counter[f"/{plugin_type}/index"] == request_count

        assert "Extractors" in result.stdout
        assert "tap-gitlab" in result.stdout
        assert "tap-mock" in result.stdout
        assert "tap-mock, variants: meltano (default), singer-io" in result.stdout

        assert "Loaders" in result.stdout
        assert "target-jsonl" in result.stdout
        assert "target-mock" in result.stdout

    def test_discover_extractors(
        self,
        project: Project,
        cli_runner,
        hub_request_counter: Counter,
    ):
        adapter = project.hub_service.session.get_adapter(
            project.hub_service.hub_api_url,
        )

        with mock.patch("requests.adapters.HTTPAdapter.send", adapter.send):
            result = cli_runner.invoke(cli, ["discover", "extractors"])

        assert_cli_runner(result)

        assert hub_request_counter["/extractors/index"] == 1
        assert hub_request_counter["/loaders/index"] == 0

        assert "Extractors" in result.stdout
        assert "tap-gitlab" in result.stdout
        assert "tap-mock" in result.stdout

        assert "Loaders" not in result.stdout



from __future__ import annotations

import pytest

from asserts import assert_cli_runner
from meltano.cli import cli


class TestAutoInstall:
    @pytest.mark.slow()
    @pytest.mark.skip(reason="Deadlock on docker, have to investigate further.")
    @pytest.mark.usefixtures("project")
    def test_carbon_intensity_sqlite(self, cli_runner):
        # Directly run `meltano add loader target-sqlite` in a new project
        # So that the extractor and the loader are auto installed by meltano
        cli_args = ["elt", "tap-carbon-intensity", "target-sqlite"]
        result = cli_runner.invoke(cli, cli_args)

        assert_cli_runner(result)



"""Project plugin service."""


from __future__ import annotations

import enum
import sys
import typing as t
from contextlib import contextmanager, suppress

import structlog

from meltano.core.environment import EnvironmentPluginConfig
from meltano.core.error import MeltanoError
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.base import VariantNotFoundError
from meltano.core.plugin.error import PluginNotFoundError, PluginParentNotFoundError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_discovery_service import (
    LockedDefinitionService,
    PluginDiscoveryService,
)
from meltano.core.plugin_lock_service import PluginLockService
from meltano.core.settings_service import FeatureFlags

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from cached_property import cached_property

if t.TYPE_CHECKING:
    from meltano.core.project import Project

logger = structlog.stdlib.get_logger(__name__)


class DefinitionSource(enum.Flag):
    """The source of a plugin definition."""

    NONE = 0
    DISCOVERY = enum.auto()
    HUB = enum.auto()
    CUSTOM = enum.auto()
    LOCKFILE = enum.auto()
    INHERITED = enum.auto()

    ANY = DISCOVERY | HUB | CUSTOM | LOCKFILE | INHERITED
    LOCAL = ~HUB


class PluginAlreadyAddedException(Exception):
    """Raised when a plugin is already added to the project."""

    def __init__(self, plugin: PluginRef, new_plugin: PluginRef):
        """Create a new Plugin Already Added Exception.

        Args:
            plugin: The plugin that was already added.
            new_plugin: The plugin that was attempted to be added.
        """
        self.plugin = plugin
        self.new_plugin = new_plugin
        super().__init__()


class PluginDefinitionNotFoundError(MeltanoError):
    """Raised when no plugin definition is found."""

    def __init__(
        self,
        plugin: ProjectPlugin,
        error: Exception | None,
        source: DefinitionSource,
    ):
        """Initialize a new error.

        Args:
            plugin: The plugin that was not found.
            error: The error that was raised, if any.
            source: The sources searched for the plugin.
        """
        reason = (
            str(error)
            if error
            else f"No definition found for {plugin.type.descriptor} {plugin.name}"
        )
        instruction = None

        if DefinitionSource.HUB in source:
            instruction = (
                f"Run `meltano discover {plugin.type.descriptor}` to explore available "
                "plugins"
            )
        else:
            instruction = (
                "Try running `meltano lock --update --all` to ensure your plugins are "
                "up to date"
            )

        super().__init__(reason=reason, instruction=instruction)


class ProjectPluginsService:  # noqa: WPS214, WPS230 (too many methods, attributes)
    """Project Plugins Service."""

    def __init__(self, project: Project):
        """Create a new Project Plugins Service.

        Args:
            project: The Meltano project.
        """
        self.project = project
        self.discovery_service = PluginDiscoveryService(project)
        self.lock_service = PluginLockService(project)
        self.locked_definition_service = LockedDefinitionService(project)
        with self.project.settings.feature_flag(
            FeatureFlags.PLUGIN_LOCKS_REQUIRED.value,
            raise_error=False,
        ) as flag:
            if flag:
                self._prefer_source = DefinitionSource.LOCAL
            else:
                self._prefer_source = DefinitionSource.ANY

    @cached_property
    def current_plugins(self):
        """Return the current plugins.

        Returns:
            The current plugins.
        """
        return self.project.config_service.current_meltano_yml.plugins

    @contextmanager
    def update_plugins(self):
        """Update the current plugins.

        Yields:
            The current plugins.
        """
        with self.project.config_service.update_meltano_yml() as meltano_yml:
            yield meltano_yml.plugins

    def add_to_file(self, plugin: ProjectPlugin):
        """Add plugin to `meltano.yml`.

        Args:
            plugin: The plugin to add.

        Raises:
            PluginAlreadyAddedException: If the plugin is already added.

        Returns:
            The added plugin.
        """
        # FIXME: `should_add_to_file` is a method from `BasePlugin`, which is
        #        not a subclass of `ProjectPlugin`. I've left this call to it
        #        in-place because I'm worried that removing it will break stuff
        #        that relies on it, but something is definitely wrong here.
        #        We default to `True` for `ProjectPlugin` objects.
        if not getattr(plugin, "should_add_to_file", lambda: True)():
            return plugin

        with suppress(PluginNotFoundError):
            existing_plugin = self.get_plugin(plugin)
            raise PluginAlreadyAddedException(existing_plugin, plugin)

        with self.update_plugins() as plugins:
            if plugin.type not in plugins:
                plugins[plugin.type] = []

            plugins[plugin.type].append(plugin)

        return plugin

    def remove_from_file(self, plugin: ProjectPlugin):
        """Remove plugin from `meltano.yml`.

        Args:
            plugin: The plugin to remove.

        Returns:
            The removed plugin.
        """
        # Will raise if the plugin isn't actually in the file
        self.get_plugin(plugin)

        with self.update_plugins() as plugins:
            plugins[plugin.type].remove(plugin)

        return plugin

    def has_plugin(self, plugin_name: str) -> bool:
        """Check if plugin exists for the given name.

        Args:
            plugin_name: The name of the plugin to check for.

        Returns:
            True if the plugin exists, False otherwise.
        """
        try:
            self.find_plugin(plugin_name)
            return True
        except PluginNotFoundError:
            return False

    def find_plugin(
        self,
        plugin_name: str,
        plugin_type: PluginType | None = None,
        invokable=None,
        configurable=None,
    ) -> ProjectPlugin:
        """
        Find a plugin.

        Args:
            plugin_name: The name of the plugin to find.
            plugin_type: Optionally the type of plugin.
            invokable: Optionally limit the search to invokable plugins.
            configurable: Optionally limit the search to configurable plugins.

        Returns:
            The plugin.

        Raises:
            PluginNotFoundError: If the plugin is not found.
        """
        if "@" in plugin_name:
            plugin_name, profile_name = plugin_name.split("@", 2)
            logger.warning(
                "Plugin configuration profiles are no longer supported, "
                f"ignoring `@{profile_name}` in plugin name.",
            )

        try:
            plugin = next(
                plugin
                for plugin in self.plugins(ensure_parent=False)
                if (
                    plugin.name == plugin_name  # noqa: WPS222 (with too much logic)
                    and (plugin_type is None or plugin.type == plugin_type)
                    and (
                        invokable is None
                        or self.ensure_parent(plugin).is_invokable() == invokable
                    )
                    and (
                        configurable is None
                        or self.ensure_parent(plugin).is_configurable() == configurable
                    )
                )
            )

            return self.ensure_parent(plugin)
        except StopIteration as stop:
            raise PluginNotFoundError(
                PluginRef(plugin_type, plugin_name) if plugin_type else plugin_name,
            ) from stop

    def find_plugin_by_namespace(
        self,
        plugin_type: PluginType,
        namespace: str,
    ) -> ProjectPlugin:
        """
        Find a plugin based on its PluginType and namespace.

        For example, PluginType.EXTRACTORS and namespace tap_custom
        will return the extractor for the tap-custom plugin.

        Args:
            plugin_type: The type of plugin to find.
            namespace: The namespace of the plugin.

        Returns:
            The plugin if found.

        Raises:
            PluginNotFoundError: If no plugin is found.
        """
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if plugin.namespace == namespace and plugin_type == plugin.type
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def find_plugins_by_mapping_name(self, mapping_name: str) -> list[ProjectPlugin]:
        """Find plugins with the specified mapping name in their mappings config.

        Args:
            mapping_name: The name of the mapping to find.

        Returns:
            The mapping plugins with the specified mapping name.

        Raises:
            PluginNotFoundError: If no mapper plugin with the specified mapping
                name is found.
        """
        found: list[ProjectPlugin] = []
        for plugin in self.get_plugins_of_type(plugin_type=PluginType.MAPPERS):
            if plugin.extra_config.get("_mapping_name") == mapping_name:
                found.append(plugin)
        if not found:
            raise PluginNotFoundError(mapping_name)
        return found

    def get_plugin(self, plugin_ref: PluginRef) -> ProjectPlugin:
        """Get a plugin using its PluginRef.

        Args:
            plugin_ref: The plugin reference to use.

        Returns:
            The plugin if found.

        Raises:
            PluginNotFoundError: If the plugin is not found.
        """
        try:
            plugin = next(
                plugin
                for plugin in self.plugins(ensure_parent=False)
                if plugin == plugin_ref
            )

            return self.ensure_parent(plugin)
        except StopIteration as stop:
            raise PluginNotFoundError(plugin_ref) from stop

    def get_plugins_of_type(
        self,
        plugin_type: PluginType,
        ensure_parent=True,
    ) -> list[ProjectPlugin]:
        """Return plugins of specified type.

        Args:
            plugin_type: The type of the plugins to return.
            ensure_parent: If True, ensure that plugin has a parent plugin set.

        Returns:
            A list of plugins of the specified plugin type.
        """
        plugins = self.current_plugins[plugin_type]

        if ensure_parent:
            for plugin in plugins:
                self.ensure_parent(plugin)

        return plugins

    def plugins_by_type(self, ensure_parent=True):
        """Return plugins grouped by type.

        Args:
            ensure_parent: If True, ensure that plugin has a parent plugin set.

        Returns:
            A dict of plugins grouped by type.
        """
        return {
            plugin_type: self.get_plugins_of_type(
                plugin_type,
                ensure_parent=ensure_parent,
            )
            for plugin_type in PluginType
        }

    def plugins(self, ensure_parent=True) -> t.Generator[ProjectPlugin, None, None]:
        """Return all plugins.

        Args:
            ensure_parent: If True, ensure that plugin has a parent plugin set.

        Yields:
            Plugins.
        """
        yield from (
            plugin
            for _, plugins in self.plugins_by_type(ensure_parent=ensure_parent).items()
            for plugin in plugins
        )

    def update_plugin(self, plugin: ProjectPlugin):
        """Update a plugin.

        Args:
            plugin: The plugin to update.

        Returns:
            The outdated plugin.
        """
        with self.update_plugins() as plugins:
            # find the proper plugin to update
            idx, outdated = next(
                (idx, plg)
                for idx, plg in enumerate(plugins[plugin.type])
                if plg == plugin
            )

            plugins[plugin.type][idx] = plugin

            return outdated

    def update_environment_plugin(self, plugin: EnvironmentPluginConfig):
        """Update a plugin configuration inside a Meltano environment.

        Args:
            plugin: The plugin configuration to update.
        """
        with self.project.config_service.update_active_environment() as environment:
            environment.config.plugins.setdefault(plugin.type, [])

            # find the proper plugin to update
            p_idx, p_outdated = next(
                (
                    (idx, plg)
                    for idx, plg in enumerate(environment.config.plugins[plugin.type])
                    if plg == plugin
                ),
                (None, None),
            )

            if p_idx is None:
                environment.config.plugins.setdefault(plugin.type, [])
                environment.config.plugins[plugin.type].append(plugin)
            else:
                environment.config.plugins[plugin.type][p_idx] = plugin

    def _get_parent_from_discovery(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Get the parent plugin from discovery.yml.

        Args:
            plugin: The plugin to get the parent of.

        Returns:
            The parent plugin.

        Raises:
            PluginParentNotFoundError: If the parent plugin is not found.
        """
        try:
            return self.discovery_service.get_base_plugin(plugin)
        except (PluginNotFoundError, VariantNotFoundError) as err:
            if plugin.inherit_from:
                raise PluginParentNotFoundError(plugin, err) from err
            raise

    def _get_parent_from_hub(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Get the parent plugin from the hub.

        Args:
            plugin: The plugin to get the parent of.

        Returns:
            The parent plugin.

        Raises:
            PluginParentNotFoundError: If the parent plugin is not found.
        """
        try:
            return self.project.hub_service.get_base_plugin(
                plugin,
                variant_name=plugin.variant,
            )
        except PluginNotFoundError as err:
            if plugin.inherit_from:
                raise PluginParentNotFoundError(plugin, err) from err
            raise

    def find_parent(
        self,
        plugin: ProjectPlugin,
    ) -> tuple[ProjectPlugin, DefinitionSource]:
        """Find the parent plugin of a plugin.

        Args:
            plugin: The plugin to find the parent of.

        Returns:
            The parent plugin and the source of the parent.

        Raises:
            PluginDefinitionNotFoundError: If the parent plugin is not found.
        """
        error = None
        if (
            plugin.inherit_from
            and not plugin.is_variant_set
            and DefinitionSource.INHERITED in self._prefer_source
        ):
            try:
                return (
                    self.find_plugin(
                        plugin_type=plugin.type,
                        plugin_name=plugin.inherit_from,
                    ),
                    DefinitionSource.INHERITED,
                )
            except PluginNotFoundError as inherited_exc:
                error = inherited_exc

        if DefinitionSource.LOCKFILE in self._prefer_source:
            try:
                return (
                    self.locked_definition_service.get_base_plugin(
                        plugin,
                        variant_name=plugin.variant,
                    ),
                    DefinitionSource.LOCKFILE,
                )
            except PluginNotFoundError as lockfile_exc:
                error = lockfile_exc

        if DefinitionSource.DISCOVERY in self._prefer_source:
            try:
                return (
                    self._get_parent_from_discovery(plugin),
                    DefinitionSource.DISCOVERY,
                )
            except Exception as discovery_exc:
                error = discovery_exc

        if DefinitionSource.HUB in self._prefer_source:
            try:
                return (self._get_parent_from_hub(plugin), DefinitionSource.HUB)
            except Exception as hub_exc:
                error = hub_exc

        raise PluginDefinitionNotFoundError(
            plugin,
            error,
            self._prefer_source,
        ) from error

    def get_parent(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Get plugin's parent plugin.

        Args:
            plugin: The plugin to get the parent of.

        Returns:
            The parent plugin or None if the plugin has no parent.
        """
        parent, source = self.find_parent(plugin)

        logger.debug(
            "Found plugin parent",
            plugin=plugin.name,
            parent=parent.name,
            source=source,
        )
        return parent

    def ensure_parent(self, plugin: ProjectPlugin) -> ProjectPlugin:
        """Ensure that plugin has a parent set.

        Args:
            plugin: To set the parent of if necessary.

        Returns:
            The plugin (updated if necessary).
        """
        if not plugin.parent:
            plugin.parent = self.get_parent(plugin)

        return plugin

    def get_transformer(self) -> ProjectPlugin:
        """Get first available Transformer plugin.

        Raises:
            PluginNotFoundError: If there is no transformer.

        Returns:
            First available transformer plugin.
        """
        transformer = next(
            iter(self.get_plugins_of_type(plugin_type=PluginType.TRANSFORMERS)),
            None,
        )
        if not transformer:
            raise PluginNotFoundError("No Plugin of type Transformer found.")
        return transformer

    @contextmanager
    def use_preferred_source(self, source: DefinitionSource) -> None:
        """Prefer a source of definition.

        Args:
            source: The source to prefer.

        Yields:
            None.
        """
        previous = self._prefer_source
        self._prefer_source = source
        yield
        self._prefer_source = previous
